searchData={"items":[{"type":"module","title":"Ecto.Adapters.FoundationDB","doc":"([Hex.pm](https://hex.pm/packages/ecto_foundationdb) | [GitHub](https://github.com/foundationdb-beam/ecto_foundationdb))\n\nAdapter module for [FoundationDB](https://www.foundationdb.org).\n\nIt uses [erlfdb](https://github.com/foundationdb-beam/erlfdb) for communicating with the database,\nand implements an Ecto-compatible Layer on top of FDB's key-value store.","ref":"Ecto.Adapters.FoundationDB.html"},{"type":"module","title":"Installation - Ecto.Adapters.FoundationDB","doc":"Install the latest stable release of FoundationDB from the\n[official FoundationDB Releases](https://github.com/apple/foundationdb/releases).\n\nThe `foundationdb-server` package is required on any system that will be running\na FoundationDB server instance. For example, it's common to\nrun the `foundationdb-server` on your development machine and on managed\ninstances running a FoundationDB cluster, but not for your stateless Elixir\napplication server in production.\n\n`foundationdb-clients` is always required.\n\nInclude `:ecto_foundationdb` in your list of dependencies in `mix.exs`:\n\n```elixir\ndefp deps do\n  [\n    {:ecto_foundationdb, \"~> 0.5\"}\n  ]\nend\n```","ref":"Ecto.Adapters.FoundationDB.html#module-installation"},{"type":"module","title":"Usage - Ecto.Adapters.FoundationDB","doc":"Define this module in your `MyApp` application:\n\n```elixir\ndefmodule MyApp.Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n\n  @impl true\n  def migrations(), do: []\nend\n```\n\nEdit your `config.exs` file to include the following:\n\n```elixir\nconfig :my_app, MyApp.Repo,\n  cluster_file: \"/etc/foundationdb/fdb.cluster\"\n```\n\nIn your app's supervisor, you can start the repo as follows:\n\n```elixir\nchildren = [\n  ...\n  MyApp.Repo,\n  ...\n]\n\nSupervisor.start_link(...)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-usage"},{"type":"module","title":"Tenants - Ecto.Adapters.FoundationDB","doc":"EctoFoundationDB requires your application to use multitenancy via `EctoFoundationDB.Tenant`.\nWe strongly encourage your application to decide on a multitenancy strategy that makes sense\nfor your problem domain (or simply use `\"default\"` to get started).\n\nOnce you have opened a tenant, you'll use the Ecto `:prefix` option to provide the tenant\nto each Repo operation.\n\nRetrieving data from a tenant yields no execution overhead. In effect, each tenant can be\nviewed as a logically partitioned database. See `EctoFoundationDB.Tenant` for a discussion\non the mechanism used to achieve this in FDB.","ref":"Ecto.Adapters.FoundationDB.html#module-tenants"},{"type":"module","title":"Creating a schema - Ecto.Adapters.FoundationDB","doc":"This is a standard `Ecto.Schema` that will be used in the examples of this documentation.\n\n```elixir\ndefmodule User do\n  use Ecto.Schema\n  @primary_key {:id, :binary_id, autogenerate: true}\n  schema \"users\" do\n    field(:name, :string)\n    field(:department, :string)\n    field(:balance, :integer)\n    timestamps()\n  end\n  # ...\nend\n```","ref":"Ecto.Adapters.FoundationDB.html#module-creating-a-schema"},{"type":"module","title":"Opening a tenant - Ecto.Adapters.FoundationDB","doc":"See `EctoFoundationDB.Tenant` for more details.\n\n```elixir\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(MyApp.Repo, \"some-org\")\n```","ref":"Ecto.Adapters.FoundationDB.html#module-opening-a-tenant"},{"type":"module","title":"Inserting a new struct - Ecto.Adapters.FoundationDB","doc":"The Tenant is stored on the struct as metadata, so that when passing structs\nto Repo functions, the `:prefix` option is not required.\n\n```elixir\nalice = %User{name: \"Alice\", department: \"Engineering\"}\n       |> FoundationDB.usetenant(tenant)\n\nalice = MyApp.Repo.insert!(alice)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-inserting-a-new-struct"},{"type":"module","title":"Querying for a struct using the primary key - Ecto.Adapters.FoundationDB","doc":"The `:prefix` option is required. The struct returned will have the tenant\nin the metadata, like above.\n\n```elixir\nMyApp.Repo.get!(User, alice.id, prefix: tenant)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-querying-for-a-struct-using-the-primary-key"},{"type":"module","title":"Transaction basics - Ecto.Adapters.FoundationDB","doc":"Here's a simple example of transaction usage. Notice that the calls to Repo inside the transaction function\ndo not include the `:prefix` option. See [Transactions](#module-transactions) for more information.\n\n```elixir\nMyApp.Repo.transactional(tenant, fn ->\n  alice = MyApp.Repo.get!(User, alice.id)\n  if alice.balance > 0 do\n    MyApp.Repo.update(User.changeset(alice, %{balance: alice.balance - 1}))\n  else\n    raise \"Overdraft\"\n  end\nend)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-transaction-basics"},{"type":"module","title":"Indexes and queries - Ecto.Adapters.FoundationDB","doc":"The implication of the [FoundationDB Layer Concept](https://apple.github.io/foundationdb/layer-concept.html)\nis that the manner in which data can be stored and accessed is the responsibility of the client\napplication. To achieve general purpose data storage and access patterns commonly desired\nin web applications, EctoFoundationDB provides support for indexes out of the box, and\nalso allows your application to define custom indexes.\n\nWhen you define a migration that calls `create index(User, [:department])`, a `Default` index is\ncreated on the `:department` field. The following two indexes are equivalent:\n\n```elixir\n# ...\ncreate index(User, [:department])\n# is equivalent to\ncreate index(User, [:department], options: [indexer: EctoFoundationDB.Indexer.Default]))]\n# ...\n```\n\nTimestamp indexes and multi-field indexes are supported:\n\n```elixir\ncreate index(User, [:inserted_at])\ncreate index(User, [:birthyear, :department, :birthdate])\n```\n\nIn fact, the index value can be any Elixir term. All types support Equal queries (of the form `x.field == \"value\"`).\nCertain Ecto Schema types support Between queries (of the form `x.field >= 10 and x.field < 20`).\nSee the [Data Types](#module-data-types) section for the list.\n\nSee the [Migrations](#module-migrations) section for further details about managing indexes in your application.","ref":"Ecto.Adapters.FoundationDB.html#module-indexes-and-queries"},{"type":"module","title":"Index query examples - Ecto.Adapters.FoundationDB","doc":"Retrieve all users in the Engineering department (with an Equal constraint):\n\n```elixir\nquery = from(u in User, where: u.department == ^\"Engineering\")\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nRetrieve all users with a birthyear from 1992 to 1994 (with a Between constraint):\n\n```elixir\nquery = from(u in User,\n  where: u.birthyear >= ^1992 and u.birthyear <= ^1994\n)\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nRetrieve all Engineers born in August 1992:\n\n```elixir\nquery = from(u in User,\n  where: u.birthyear == ^1992 and\n         u.department == ^\"Engineering\" and\n         u.birthdate >= ^~D[1992-08-01] and u.birthdate < ^~D[1992-09-01]\n)\nMyApp.Repo.all(query, prefix: tenant)\n```","ref":"Ecto.Adapters.FoundationDB.html#module-index-query-examples"},{"type":"module","title":"Index ordering matters - Ecto.Adapters.FoundationDB","doc":"When you create an index using multiple fields, the FDB key that stores the index will be extended with\nall the values *in the order of your defined index fields*. Because FoundationDB stores the keys themselves\nin a well-defined order, the order of the fields in your index determines the Between queries you can perform.\n\n> **_RULE:_**  For any Query, there can be 0 or 1 Between clauses, and if one exists for a field, then there must not\nexist an Equal clause for ay field that follows, according to the list of fields in the index. Moreover, the Between\nfield must be of a [Data Type](#module-data-types) supported for Between queries.\n\nAbove, we demonstrated a nontrivial query with 2 Equal constraints and 1 Between constraint. Compare that with an invalid\nquery below, which attempts a Between constraint on the `:birthyear` field followed by an Equal constraint on the `:department`\nfield.\n\n```elixir\n# Invalid query!\nfrom(u in User,\n  where: u.birthyear >= ^1992 and u.birthyear < ^1995 and\n         u.department == ^\"Engineering\"\n)\n```\n\n**Why is this query unsupported by EctoFDB?** With the index defined as-is, EctoFDB is unable to extract a minimal\ndataset from the database in one operation. Instead of over-extracting data and then applying a filter, we've chosen\nto reject the query. We hope this encourages correct index management and gives the developer confidence about the\nexpected execution time for any given call to Repo.\n\n**Yeah that's cool, but how do I run the query I want to run?** You have 2 options.\n\n1. Purposefully over-extract the data using Stream, and do the filtering in Elixir:\n\n```elixir\nfrom(u in User, where: u.birthyear >= ^1992 and u.birthyear < 1995 and)\n|> MyApp.Repo.stream(prefix: tenant)\n|> Stream.filter(&((&1).department == \"Engineering\"))\n|> Enum.to_list()\n```\n\nor\n\n2. Create the necessary index. For our example above, the new index might look like:\n\n```elixir\ncreate index(User, [:department, :birthyear])\n```\n\nWe encourage you to test execution times of your expected workload before deciding to create an index.\nRemember: since your data is already partitioned into tenants, you're effectively getting 1 index for free.","ref":"Ecto.Adapters.FoundationDB.html#module-index-ordering-matters"},{"type":"module","title":"Custom indexes - Ecto.Adapters.FoundationDB","doc":"As data is inserted, updated, deleted, and queried, the Indexer callbacks (via behaviour `EctoFoundationDB.Indexer`)\nare executed. Your Indexer module does all the work necessary to maintain the index within the transaction\nthat the data itself is being manipulated. Many different types of indexes are possible. You are in control!\n\nFor an example, please see the [Default implementation](https://github.com/foundationdb-beam/ecto_foundationdb/blob/main/lib/ecto_foundationdb/indexer/default.ex).","ref":"Ecto.Adapters.FoundationDB.html#module-custom-indexes"},{"type":"module","title":"Transactions - Ecto.Adapters.FoundationDB","doc":"All work involving FDB happens in a transaction, including each Repo function call (e.g. `Repo.get/1`). EctoFoundationDB also\nexposes the `Repo.transactional/2` function, allowing you to group operations together with transactional isolation ([ACID](https://en.wikipedia.org/wiki/ACID)\nand [globally serializable](https://en.wikipedia.org/wiki/Global_serializability))\n\n```elixir\nMyApp.Repo.transactional(tenant, fn ->\n    # Ecto work\n  end)\n```\n\nPlease read the [FoundationDB Developer Guide on transactions](https://apple.github.io/foundationdb/developer-guide.html#transaction-basics)\nfor more information about how to develop with transactions.\n\nIt's important to remember that even though the database gives ACID guarantees about the\nkeys updated in a transaction, the Elixir function passed into `Repo.transactional/2`\nmay be executed more than once when any conflicts\nare encountered. For this reason, your function must not have any side effects other than\nthe database operations. For example, do not publish any messages to other processes\n(such as `Phoenix.PubSub`) from within the transaction.\n\n```elixir\n# Do not do this!\nMyApp.Repo.transactional(tenant, fn ->\n  MyApp.Repo.insert(%User{name: \"Alice\"})\n  ...\n  Phoenix.PubSub.broadcast(\"my_topic\", \"new_user\") # Not safe! :(\nend)\n```\n\n```elixir\n# Instead, do this:\nMyApp.Repo.transactional(tenant, fn ->\n  MyApp.Repo.insert(%User{name: \"Alice\"})\n  ...\nend)\nPhoenix.PubSub.broadcast(\"my_topic\", \"new_user\") # Safe :)\n```\n\nNote: If you're looking for PubSub-like functionality pushed from the database itself, please see\n[Watches](#module-watches).","ref":"Ecto.Adapters.FoundationDB.html#module-transactions"},{"type":"module","title":"Migrations - Ecto.Adapters.FoundationDB","doc":"At first glance, EctoFoundationDB migrations may look similar to that of `:ecto_sql`,\nbut the actual execution of migrations and how your app must be configured are very\ndifferent, so please read this section in full.\n\nIf your app uses indexes on any of your schemas, you must include a module that implements\nthe `EctoFoundationDB.Migrator` behaviour. By default, this is assumed to be your Repo\nmodule, as we have demonstrated above.\n\nMigrations are only used for index management. There are no tables to add, drop, or modify.\nWhen you add a field to your Schema, any read requests on old objects will return `nil` in\nthat field.","ref":"Ecto.Adapters.FoundationDB.html#module-migrations"},{"type":"module","title":"Online migrations - Ecto.Adapters.FoundationDB","doc":"As tenants are opened during your application runtime, migrations will be executed\nautomatically in a \"lazy\" fashion. This distributes the migration across a potentially long period of time,\nas migrations will not be executed unless the tenant is opened. Specifically, a call to\n`Tenant.open/2` will block until the migration for that tenant completes. Other tenants\nwill not be affected.\n\nFor example, imagine you have 2 nodes running your application. Node 1 has opened tenants `\"a\"`\nand `\"b\"`. You deploy a new version of your app to Node 2 that includes a new index, and open\ntenant `\"a\"` there. On Node 2 only, that call to `Tenant.open/2` will block until the migration\ncompletes. All operations on Node 1 will remain working. When `Tenant.open/2` completes on Node 2\nfor tenant `\"a\"`, the index is available to all `\"a\"` tenants across all nodes. Any `\"b\"` tenants\nremain unaffected throughout this process.","ref":"Ecto.Adapters.FoundationDB.html#module-online-migrations"},{"type":"module","title":"Configuring your app - Ecto.Adapters.FoundationDB","doc":"The Migrator is a module in your application runtime that provides the full list of\nordered migrations. These are the migrations that will be executed when a tenant is opened.\nIf you leave out a migration from the list, it will not be applied.\n\nThe result of `migrations/0` will change over time to include new migrations, like so:\n\n```elixir\ndefmodule MyApp.Repo do\n  # ...\n  def migrations() do\n    [\n      {0, MyApp.IndexUserByDepartment}\n      # At a later date, we my add a new index with a corresponding addition\n      # tho the migrations list:\n      #    {1, MyApp.IndexUserByRole}\n    ]\n  end\nend\n```\n\nEach migration is contained in a separate module, much like EctoSQL's. However, the operations\nto be carried out **must be returned as a list.** For example, the creation an index\nmay look like this:\n\n```elixir\ndefmodule MyApp.IndexUserByDepartment do\n  use EctoFoundationDB.Migration\n\n  @impl true\n  def change() do\n    [\n      create index(User, [:department])\n    ]\n  end\nend\n```\n\nYour migrator and `MyApp. ` modules are part of your codebase. They must be\nincluded in your application release. This differs from traditional Ecto migrations (`:ecto_sql`), which\nare typically managed in `.exs` scripts executed outside the production application runtime.","ref":"Ecto.Adapters.FoundationDB.html#module-configuring-your-app"},{"type":"module","title":"Creating and dropping indexes - Ecto.Adapters.FoundationDB","doc":"In your Migration, to create an index use `create index(table, columns)` and to drop an index use `drop index(table, columns)`.\n\nIndexes are created and dropped in a manner that is safe for your application transactions.\n\n```elixir\ndefmodule MyApp.ExampleMigration do\n  use EctoFoundationDB.Migration\n\n  @impl true\n  def change() do\n    [\n      create index(User, [:role]),\n      drop index(User, [:department]),\n    ]\n  end\nend\n```","ref":"Ecto.Adapters.FoundationDB.html#module-creating-and-dropping-indexes"},{"type":"module","title":"Executing all migrations immediately - Ecto.Adapters.FoundationDB","doc":"Migrations can be completed in full at any time with a call to\n`EctoFoundationDB.CLI.migrate!/1`, which accepts your repo module (e.g. `MyApp.Repo`).\n\n```elixir\n# May take a while to complete\nEctoFoundationDB.CLI.migrate!(MyApp.Repo)\n```\n\nDepending on how many tenants you have in your database, and the size of the data for each tenant,\nthis operation might take a long time and make heavy use of system resources. It is safe to interrupt\nthis operation, as migrations are performed transactionally. However, if you interrupt a migration, the next\nattempt for that tenant may have a brief delay (~5 sec) while the migrator ensures that the\nprevious migration has indeed stopped executing.\n\nPlease refer to the documentation of `EctoFoundationDB.CLI` for other actions to be carried out\nby a human operator.","ref":"Ecto.Adapters.FoundationDB.html#module-executing-all-migrations-immediately"},{"type":"module","title":"Data types - Ecto.Adapters.FoundationDB","doc":"EctoFoundationDB stores your struct's data using `:erlang.term_to_binary/1`, and\nretrieves it with `:erlang.binary_to_term/1`. As such, there is no data type\nconversion between Elixir types and Database Types. Any term you put in your\nstruct will be stored and later retrieved.\n\nData types *are* used by EctoFoundationDB for the creation and querying of indexes.\nCertain Ecto types are encoded into the FDB key, which allows you to formulate\nBetween queries on indexes of these types.\n\n** Between queries are supported for these types:**\n\n - `:id`\n - `:binary_id`\n - `:integer`\n - `:float`\n - `:boolean`\n - `:string`\n - `:binary`\n - `:date`\n - `:time`\n - `:time_usec`\n - `:naive_datetime`\n - `:naive_datetime_usec`\n - `:utc_datetime`\n - `:utc_datetime_usec`\n\nSee [Indexes and Queries](#module-indexes-and-queries) for more information.","ref":"Ecto.Adapters.FoundationDB.html#module-data-types"},{"type":"module","title":"Versionstamps (autoincrement) - Ecto.Adapters.FoundationDB","doc":"FoundationDB provides a mechanism for generating an automatically increasing integer id, called a [versionstamp](https://apple.github.io/foundationdb/data-modeling.html#versionstamps).\nVersionstamps are guaranteed to be unique and increasing along with the total order of the serializable transactions.\nThey are not guaranteed to increment by exactly one.\n\nTo use Versionstamps, your Schema must have a primary key field of type `Versionstamp`, with `autogenerate: false`.\n\n```elixir\n@primary_key {:id, EctoFoundationDB.Versionstamp, autogenerate: false}\n```\n\nThe easiest way to insert such records is by using `Repo.async_insert_all/3`. This function follows a different\napproach than the other `Repo.async_*` functions: the future must be awaited *outside* of the transaction.\n\n```elixir\nalias MyApp.{Repo, Event}\nfuture = Repo.transactional(tenant, fn ->\n  Repo.async_insert_all(Event, [%Event{data: \"event_a\"}, %Event{data: \"event_b\"}])\nend)\n[event_a, event_b] = Repo.await(future)\n```\n\nThe purpose of the async/await pattern here is for the FDB client to discover the committed versionstamp, and to apply\nit to the inserted structs after the transaction is committed. Since the computation of the versionstamp is a\ncommit-time decision by the FoundationDB Server, the final value is unknowable by the client until after the transaction\nis committed.\n\nInspecting the contents of the `:id` field will reveal the structure of the versionstamp:\n\n```elixir\nevent_a.id\n# => {:versionstamp, 1111, 2222, 0}\n```\n\nThe Versionstamp can be converted to an integer using the `EctoFoundationDB.Versionstamp.to_integer/1` function.\n\n```elixir\nalias EctoFoundationDB.Versionstamp\nVersionstamp.to_integer(event_a.id)\n# => 4771854286848\n```\n\nNote:\n\n - The approach to inserting above is guaranteed to have zero conflicts with other keys in the database.\n - A group of records inserted in a single transaction **will** have versionatmps that are guaranteed to increment by 1.\n - Records inserted in different transactions **will not** have a predictable versionstamp distance from each other. That distance\n   is very unlikely to ever be 1.","ref":"Ecto.Adapters.FoundationDB.html#module-versionstamps-autoincrement"},{"type":"module","title":"Watches - Ecto.Adapters.FoundationDB","doc":"[FoundationDB Watches](https://apple.github.io/foundationdb/developer-guide.html#watches) are\nsimilar to Triggers in an RDBMS. Registering a watch on a particular key provides a guarantee[*](https://github.com/apple/foundationdb/wiki/An-Overview-how-Watches-Work)\nthat when that key in the database is changed, the client application is notified with a\npush-style notification. This registration is initiated inside a transaction, and the caller is\nprovided a `EctoFoundationDB.Future` object that is used to receive the push notification at a later time.\nThe notification is delivered directly to the Elixir process that requested it.\n\nThe EctoFDB Repo module allows you to register a watch on any struct that you have written\nto the database. The FDB watch is registered on the key that represents the Primary Write (the key-value\nthat stores all the data of your struct). When that struct changes in the database via any means,\nthe Future is resolved and the caller has the opportunity to refresh its internal state.","ref":"Ecto.Adapters.FoundationDB.html#module-watches"},{"type":"module","title":"EctoFDB Repo functions for watches - Ecto.Adapters.FoundationDB","doc":"EctoFDB adds the following functions to your Repo module. They are each discussed in detail\nbelow.\n\n- `Repo.watch/1`/`Repo.watch/2`: Registers a watch\n- `Repo.assign_ready/2`/`Repo.assign_ready/3`: Handles Future resolution via a push message\n\n### `Repo.watch/2`\n\nWith arguments `struct` and `options`, registers\na watch on the given struct and returns a Future. This Future resolves to\na value when the FDB key is set or cleared. Until then, the Future remains unresolved.\nThe Future from a watch is resolvable outside of an FDB transaction.\n\nIn addition to other standard options, a `:label` option will prime the Future for use by `assign_ready/3`.\nIt's recommended that you provide a label and use `assign_ready/3`, but not required.\n\n### `Repo.assign_ready/3`\n\nWith arguments `futures`, `ready_refs`, and\n`options`, returns a tuple containing: 1) a Keyword of structs with keys according to the `:label`\nprovided to `watch/2`, and 2) an updated list of futures that are still unresolved.\n\nIn order for the caller to acquire the `ready_refs` arguments, you must receive messages from the\nprocess mailbox of the form `{ref, :ready}`. See the example below for detail. If your process chooses\nto buffer a list `ready_refs` longer than 1, then `assign_ready/3` uses pipelining to retrieve the\nresulting structs. Each watched struct must have been registered with a unique `:label`.\n\nThe `options` are provided to `Repo.async_get/3` internally. In addition to standard options, a\n`watch?: true` option will re-register a new watch on the struct so that your calling\nprocess's event loop can always be aware of changes.","ref":"Ecto.Adapters.FoundationDB.html#module-ectofdb-repo-functions-for-watches"},{"type":"module","title":"Example - Ecto.Adapters.FoundationDB","doc":"The power of watches is best illustrated in the context of a stateful Elixir process that is\ninterested in tracking the up-to-date value for a particular schema. This could be a GenServer, LiveView,\nor some other stateful process.\n\n```elixir\ndefmodule MyAliceView do\n\n  # ...\n\n  def init(tenant) do\n    {alice, future} = MyRepo.transactional(tenant,\n                        fn ->\n                          alice = MyRepo.get_by(User, name: \"Alice\")\n                          {alice, MyRepo.watch(alice, label: :alice)}\n                        end)\n    assigns = [alice: alice]\n    {:ok, %{assigns: assigns, futures: [future]}}\n  end\n\n  # ...\n\n  def handle_info({ref, :ready}, state=%{assigns: assigns, futures: futures, tenant: tenant}) when is_reference(ref) do\n    {new_assigns, new_futures} = MyRepo.assign_ready(futures, [ref], watch?: true, prefix: tenant)\n    {:noreply, %{\n          state |\n            assigns: Map.merge(assigns, Enum.into(new_assigns, %{})),\n            futures: new_futures\n          }}\n  end\nend\n```\n\nBy handling the Future message resolution event `{ref, :ready}` in the `handle_info/2`, we never need to\ncall `Repo.await/1`. Instead, our process can do other work and react to the `:ready` event immediately\nwhen it's delivered.\n\nAnd `Repo.assign_ready/3` provides a conventional way to retrieve the result and register a new watch\nin one go, so that updates to Alice are not missed.\n\nFor more on watches, see\n\n* [Sync Engine Part I - Single Object](watches.html)\n* [Schema Metadata](#module-schema-metadata)","ref":"Ecto.Adapters.FoundationDB.html#module-example"},{"type":"module","title":"Schema Metadata - Ecto.Adapters.FoundationDB","doc":"In addition to [Default Indexes](#module-indexes-and-queries), EctoFDB provides another built-in Indexer called `EctoFoundationDB.Indexer.SchemaMetadata`.\n\nThis is an optional Indexer that will keep track of various actions for a Schema:\n\n- `inserts`: Incremented for each insert or upsert\n- `deletes`: Incremented for each delete\n- `collection`: Incremented for each insert, update, or delete\n- `updates`: Incremented for each update (via `Repo.update/*`)\n- `changes`: Incremented for each insert, upsert, delete, or update\n\nThese keys are useful for creating watches that will notify your application of those actions. For example, if you create a watch on the `inserts` key,\nyour application will be notified when a new record is inserted, and you can react however you like.\n\nThe counters are scoped to a particular tenant. If you desire a different scoping for the watches, then we recommend you split your tenant.\n\nTo create this Indexer, you use [Migrations](#module-migrations):\n\n```elixir\ndefmodule MyApp.UserSchemaMetadata do\n  @moduledoc false\n  use EctoFoundationDB.Migration\n\n  @impl true\n  def change() do\n    [create metadata(User)]\n  end\nend\n```\n\n```elixir\n  def migrations() do\n    [\n      # ...\n      {2, MyApp.UserSchemaMetadata},\n      # ...\n    ]\n  end\n```\n\nFor more on these watches, see [Sync Engine Part II - Collections](collection_syncing.html)","ref":"Ecto.Adapters.FoundationDB.html#module-schema-metadata"},{"type":"module","title":"Upserts - Ecto.Adapters.FoundationDB","doc":"The FoundationDB Adapter supports the following upsert approaches.","ref":"Ecto.Adapters.FoundationDB.html#module-upserts"},{"type":"module","title":"Upserts with `on_conflict` - Ecto.Adapters.FoundationDB","doc":"The following choices for `on_conflict` are supported, and they work in the manner\ndescribed by the official Ecto documentation.\n\n- `:raise`: The default.\n- `:nothing`\n- `:replace_all`\n- `{:replace_all_except, fields}`\n- `{:replace, fields}`","ref":"Ecto.Adapters.FoundationDB.html#module-upserts-with-on_conflict"},{"type":"module","title":"Upserts with `conflict_target` - Ecto.Adapters.FoundationDB","doc":"If you provide `conflict_target: []`, the FoundationDB Adapter will make no effort to\ninspect the existence of objects in the database before insert. When an existing\nobject is blindly overwritten, your indexes may become inconsistent, resulting in undefined\nbehavior. You should only consider its use if one of the following is true:\n\n- (a) your schema has no indexes\n- (b) you know apriori that any indexed fields are unchanged at the time of upsert, or\n- (c) you know apirori that the primary key you're inserting doesn't already exist.\n\nFor these use cases, this option may speed up the loading of initial data at scale.","ref":"Ecto.Adapters.FoundationDB.html#module-upserts-with-conflict_target"},{"type":"module","title":"Pipelining - Ecto.Adapters.FoundationDB","doc":"Pipelining is the technique of sending multiple queries to a database at the same time, and\nthen receiving the results at a later time. In doing so, you can often avoid waiting\nfor multiple network round trips. EctoFDB uses a `async`/`await` syntax to\nimplement pipelining.","ref":"Ecto.Adapters.FoundationDB.html#module-pipelining"},{"type":"module","title":"EctoFDB Repo functions for pipelining - Ecto.Adapters.FoundationDB","doc":"EctoFDB adds the following functions to your Repo module. You'll notice that there are `async_*`\nversions of all the Queryable functions on the standard `Ecto.Repo`. These have the same arguments\nas their counterparts, and they each return a single `EctoFoundationDB.Future`.\n\n- `Repo.async_get/2` / `Repo.async_get/3`\n- `Repo.async_get!/2` / `Repo.async_get!/3`\n- `Repo.async_get_by/2` / `Repo.async_get_by/3`\n- `Repo.async_get_by!/2` / `Repo.async_get_by!/3`\n- `Repo.async_one/1` / `Repo.async_one/2`\n- `Repo.async_one!/1` / `Repo.async_one!/2`\n- `Repo.async_all/1` / `Repo.async_all/2`\n- `Repo.await/1`: The results of the above functions are provided to `Repo.await/1` to\n  resolve the Futures. It accepts a single Future or a list of Futures.","ref":"Ecto.Adapters.FoundationDB.html#module-ectofdb-repo-functions-for-pipelining"},{"type":"module","title":"Example - Ecto.Adapters.FoundationDB","doc":"Consider this example. Our transaction below has 2 network round trips in serial.\nThe \"Alice\" User is retrieved, and then the \"Bob\" user is retrieved. Finally, the\nlist is constructed and the transaction ends.\n\n```elixir\nresult = MyApp.Repo.transactional(tenant, fn ->\n  [\n    MyApp.Repo.get_by(User, name: \"Alice\"),\n    MyApp.Repo.get_by(User, name: \"Bob\")\n  ]\nend)\n\n[alice, bob] = result\n```\n\nWith pipelining, we can issue 2 `async_get_by` queries without waiting for their result,\nand then only when we need the data, we `await` the result. Notice that the list of futures\nis constructed, and then we `await` on them.\n\n```elixir\nresult = MyApp.Repo.transactional(tenant, fn ->\n  futures = [\n    MyApp.Repo.async_get_by(User, name: \"Alice\"),\n    MyApp.Repo.async_get_by(User, name: \"Bob\")\n  ]\n\n  MyApp.Repo.await(futures)\nend)\n\n[alice, bob] = result\n```\n\nConceptually, only 1 network round-trip is required, so this will be faster than\nwaiting on each in sequence.","ref":"Ecto.Adapters.FoundationDB.html#module-example"},{"type":"module","title":"Pipelining on range queries (interleaving waits) - Ecto.Adapters.FoundationDB","doc":"The functions `Repo.async_all/1` and `Repo.async_all/2` work a little bit different on the await. When the DB must\nreturn a list of results, if that list ends up being larger than `:erlfdb`'s configured `:target_bytes`,\nthen the result set is split into multiple round-trips to the database. If your application awaits multiple\nsuch queries, then `:erlfdb`'s \"interleaving waits\" feature is used.\n\nConsider this example, where we're retrieving all Users and all Posts in the same transaction.\n\n```elixir\nMyApp.Repo.transactional(tenant, fn ->\n  futures = [\n    MyApp.Repo.async_all(User),\n    MyApp.Repo.async_all(Post)\n  ]\n\n  MyApp.Repo.await(futures)\nend)\n```\n\nEach query individually executed would be multiple round-trips to the database as pages of results are retrieved.\n\n```mermaid\nsequenceDiagram\n    Client->>Server: get_range(User)\n    Server-->>Client: {100 users, continuation_token}\n    Client->>Server: get_range(User, continuation_token)\n    Server-->>Client: {10 users, done}\n```\n\nWith `:erlfdb`'s interleaving waits, we send as many get_range requests as possible with pipelining\nuntil all are exhausted.\n\n```mermaid\nsequenceDiagram\n    Client->>Server: get_range(User)\n    Client->>Server: get_range(Post)\n    Server-->>Client: {100 users, continuation_token}\n    Server-->>Client: {100 posts, continuation_token}\n    Client->>Server: get_range(User, continuation_token)\n    Client->>Server: get_range(Post, continuation_token)\n    Server-->>Client: {10 users, done}\n    Server-->>Client: {100 posts, continuation_token}\n    Client->>Server: get_range(Post, continuation_token)\n    Server-->>Client: {2 posts, done}\n```\n\nThis happens automatically when you include the list of futures to your `Repo.await/1`.","ref":"Ecto.Adapters.FoundationDB.html#module-pipelining-on-range-queries-interleaving-waits"},{"type":"module","title":"Standard options - Ecto.Adapters.FoundationDB","doc":"* `:cluster_file` - The path to the fdb.cluster file. The default is\n    `\"/etc/foundationdb/fdb.cluster\"`.\n  * `:migrator` - A module that implements the `EctoFoundationDB.Migrator`\n    behaviour. Required when using any indexes. Defaults to `nil`. When `nil`,\n    your Repo module is assumed to be the Migrator.","ref":"Ecto.Adapters.FoundationDB.html#module-standard-options"},{"type":"module","title":"Advanced options - Ecto.Adapters.FoundationDB","doc":"* `:storage_id` - All tenants created by this adapter are prefixed with\n    this string. This allows multiple configurations to operate on the\n    same FoundationDB cluster independently. Defaults to\n    `\"Ecto.Adapters.FoundationDB\"`.\n  * `:open_db` - 1-arity function accepting the Repo module. Used for opening a reference to the\n     FoundationDB cluster. Defaults to `&EctoFoundationDB.Database.open/1`. When\n     using `EctoFoundationDB.Sandbox`, you should set this option to `Sandbox.open_db/1`.\n  * `:migration_step` - The maximum number of keys to process in a single transaction\n    when running migrations. Defaults to `1000`. If you use a number that is\n    too large, the FDB transactions run by the Migrator will fail.\n  * `:metadata_cache` - When set to `:enabled`, the Ecto ets cache is used to store the\n    available indexes per tenant. This speeds up all database operations.\n    Defaults to `:enabled`.\n  * `:max_single_value_size` - Number of Bytes above which an encoded struct will\n    be split into multiple FDB key-value pairs upon insert and update. Defaults to `100_000`.\n    Any value beyond `100_000` is unsupported by FoundationDB. You probably do not\n    want to change this value. Also see `:max_value_size`.\n  * `:max_value_size` - Number of Bytes above which an encoded struct will be\n    rejected by EctoFDB upon insert and update. Defaults to `:infinity`, which\n    means that a single encoded struct of any size can be written to any number of\n    keys. Please remember that FDB's transaction size limit of 10MB still applies.\n    The maximum number of key-value pairs for a given encoded struct, not counting\n    indexes, is computed by `1 + ceil(:max_value_size / :max_single_value_size)`.","ref":"Ecto.Adapters.FoundationDB.html#module-advanced-options"},{"type":"module","title":"Optimizing for throughput and latency - Ecto.Adapters.FoundationDB","doc":"Pipelining operations inside your transactions can offer significant speed-ups, and you should\nalso consider parallelizing multiple transactions. Both techniques are useful in different\ncircumstances.\n\n- Pipelining: The transaction can contain arbitrary logic and remain ACID compliant. If you make\n  smart use of pipelining, the latency of execution of that logic is kept to a minimum.\n- Parallel transactions: Each transaction is internally ACID compliant, but any 2 given\n  transactions may conflict if they're accessing the same keys. When transactions do not conflict,\n  throughput goes up.\n\nIf you're attempting to get the maximum throughput on data loading, you'll probably want to\nmake use of both!\n\nRead more at [FDB Developer Guide | Loading Data](https://apple.github.io/foundationdb/developer-guide.html#loading-data).","ref":"Ecto.Adapters.FoundationDB.html#module-optimizing-for-throughput-and-latency"},{"type":"module","title":"Limitations and caveats - Ecto.Adapters.FoundationDB","doc":"","ref":"Ecto.Adapters.FoundationDB.html#module-limitations-and-caveats"},{"type":"module","title":"FoundationDB - Ecto.Adapters.FoundationDB","doc":"Please read [FoundationDB Known Limitations](https://apple.github.io/foundationdb/known-limitations.html).","ref":"Ecto.Adapters.FoundationDB.html#module-foundationdb"},{"type":"module","title":"Layer - Ecto.Adapters.FoundationDB","doc":"EctoFoundationDB implements a specific Layer on the FoundationDB\nkey-value store. This Layer is intended to be generally useful, but you\nmay not find it suitable for your workloads. The Layer is documented in\ndetail at `EctoFoundationDB.Layer`. This project does not support\nplugging in other Layers.","ref":"Ecto.Adapters.FoundationDB.html#module-layer"},{"type":"module","title":"Tenants - Ecto.Adapters.FoundationDB","doc":"* The use of a tenant implies an ownership relationship between the tenant and the data.\n   It's up to you how you use this relationship in your application.\n\n * Failure to specify a tenant will result in a raised exception at runtime.","ref":"Ecto.Adapters.FoundationDB.html#module-tenants"},{"type":"module","title":"Migrations - Ecto.Adapters.FoundationDB","doc":"The following are not supported:\n\n* Renaming a table\n* Renaming a field\n* Deleting a field [GitHub #32](https://github.com/foundationdb-beam/ecto_foundationdb/issues/32)\n* Dropping an index\n* Moving down in migration versions (rollback)\n\nFinally, do not use the  Ecto Mix tasks:\n\n```elixir\n# These commands are not supported. Do not use them with :ecto_foundationdb!\n#    mix ecto.create\n#    mix ecto.drop\n#    mix ecto.migrate\n#    mix ecto.gen.migration\n#    mix ecto.rollback\n#    mix ecto.migrations\n```","ref":"Ecto.Adapters.FoundationDB.html#module-migrations"},{"type":"module","title":"Key and value size - Ecto.Adapters.FoundationDB","doc":"[FoundationDB has limits on key and value size](https://apple.github.io/foundationdb/known-limitations.html#large-keys-and-values)\nPlease be aware of these limitations as you develop. EctoFDB does support splitting an individual struct\nacross multiple keys in the database. This is automatic and doesn't require any additional input from the developer. The\n[options](#module-advanced-options) `:max_single_value_size` and `:max_value_size` allow you to tweak this feature.","ref":"Ecto.Adapters.FoundationDB.html#module-key-and-value-size"},{"type":"module","title":"Ecto queries - Ecto.Adapters.FoundationDB","doc":"You'll find that most queries outside of those detailed in the documentation fail with\nan `EctoFoundationDB.Exception.Unsupported` exception. The FoundationDB Layer concept\nprecludes complex queries from being executed within the database. Therefore, it only makes sense\nto implement a limited set of query types -- specifically those that do have optimized database\nquery semantics. All other filtering, aggregation, and grouping must be done by your Elixir code.\n\nIn other words, you'll be writing less SQL, and more Elixir.\n\nRefer to the integration tests for a collection of queries that is known to work.","ref":"Ecto.Adapters.FoundationDB.html#module-ecto-queries"},{"type":"module","title":"Other Ecto features - Ecto.Adapters.FoundationDB","doc":"As you test the boundaries of EctoFoundationDB, you'll find that many other Ecto features just plain\ndon't work with this adapter. Ecto has a lot of features, and this adapter is in active development.\nPlease let us know what you find!","ref":"Ecto.Adapters.FoundationDB.html#module-other-ecto-features"},{"type":"function","title":"Ecto.Adapters.FoundationDB.db/1","doc":"","ref":"Ecto.Adapters.FoundationDB.html#db/1"},{"type":"function","title":"Ecto.Adapters.FoundationDB.transactional/2","doc":"Executes the given function in a transaction on the database.\n\nIf you provide an arity-0 function, your function will be executed in\na newly spawned process. This is to ensure that EctoFoundationDB can\nsafely manage the process dictionary.\n\nPlease be aware of the\n[limitations that FoundationDB](https://apple.github.io/foundationdb/developer-guide.html#transaction-basics)\nimposes on transactions.\n\nFor example, a transaction must complete\n[within 5 seconds](https://apple.github.io/foundationdb/developer-guide.html#long-running-transactions).","ref":"Ecto.Adapters.FoundationDB.html#transactional/2"},{"type":"function","title":"Ecto.Adapters.FoundationDB.usetenant/2","doc":"","ref":"Ecto.Adapters.FoundationDB.html#usetenant/2"},{"type":"module","title":"EctoFoundationDB.CLI","doc":"This module provides functions that are to be run by an operator, and not\npart of application code.","ref":"EctoFoundationDB.CLI.html"},{"type":"function","title":"EctoFoundationDB.CLI.copy_field!/5","doc":"For each schema item in the database, copies the contents of one field to another.","ref":"EctoFoundationDB.CLI.html#copy_field!/5"},{"type":"function","title":"Arguments - EctoFoundationDB.CLI.copy_field!/5","doc":"- `repo`: The repository to use.\n- `schema`: The schema to copy from.\n- `from`: The field to copy from.\n- `to`: The field to copy to.\n- `options`: Options for the operation.","ref":"EctoFoundationDB.CLI.html#copy_field!/5-arguments"},{"type":"function","title":"Requirements - EctoFoundationDB.CLI.copy_field!/5","doc":"- Both fields must exist on the Schema.","ref":"EctoFoundationDB.CLI.html#copy_field!/5-requirements"},{"type":"function","title":"Options - EctoFoundationDB.CLI.copy_field!/5","doc":"- `:prefix`: If provided, only this tenant will be processed. If not provided, then all tenants concurrently.\n- `:max_rows`: Maximum number of items to process in a single transaction.\n\nAlso supports options from `run_concurrently_for_all_tenants!/3` and `EctoFoundationDB.Tenant.open/3`.","ref":"EctoFoundationDB.CLI.html#copy_field!/5-options"},{"type":"function","title":"EctoFoundationDB.CLI.copy_fields!/4","doc":"For each schema item in the database, copies the contents of a set of fields to another set of fields.","ref":"EctoFoundationDB.CLI.html#copy_fields!/4"},{"type":"function","title":"Arguments - EctoFoundationDB.CLI.copy_fields!/4","doc":"- `repo`: The repository to use.\n- `schema`: The schema to copy from.\n- `copies`: A Keyword list of fields to copy from and to.\n- `options`: Options for the operation.","ref":"EctoFoundationDB.CLI.html#copy_fields!/4-arguments"},{"type":"function","title":"Requirements - EctoFoundationDB.CLI.copy_fields!/4","doc":"- All fields must exist on the Schema.","ref":"EctoFoundationDB.CLI.html#copy_fields!/4-requirements"},{"type":"function","title":"Options - EctoFoundationDB.CLI.copy_fields!/4","doc":"- `:prefix`: If provided, only this tenant will be processed. If not provided, then all tenants concurrently.\n- `:max_rows`: Maximum number of items to process in a single transaction.\n\nAlso supports options from `run_concurrently_for_all_tenants!/3` and `EctoFoundationDB.Tenant.open/3`.","ref":"EctoFoundationDB.CLI.html#copy_fields!/4-options"},{"type":"function","title":"EctoFoundationDB.CLI.delete_field!/4","doc":"For each schema item in the database, deletes the given field.","ref":"EctoFoundationDB.CLI.html#delete_field!/4"},{"type":"function","title":"Arguments - EctoFoundationDB.CLI.delete_field!/4","doc":"- `repo`: The repository to use.\n- `schema`: The schema to copy from.\n- `field`: The field to delete.\n- `options`: Options for the operation.","ref":"EctoFoundationDB.CLI.html#delete_field!/4-arguments"},{"type":"function","title":"Requirements - EctoFoundationDB.CLI.delete_field!/4","doc":"- The field must not exist on the Schema.\n- The field must not be used in any registered metadata (indexes).","ref":"EctoFoundationDB.CLI.html#delete_field!/4-requirements"},{"type":"function","title":"Options - EctoFoundationDB.CLI.delete_field!/4","doc":"- `:prefix`: If provided, only this tenant will be processed. If not provided, then all tenants concurrently.\n- `:max_rows`: Maximum number of items to process in a single transaction.\n\nAlso supports options from `run_concurrently_for_all_tenants!/3` and `EctoFoundationDB.Tenant.open/3`.","ref":"EctoFoundationDB.CLI.html#delete_field!/4-options"},{"type":"function","title":"EctoFoundationDB.CLI.delete_fields!/4","doc":"For each schema item in the database, deletes the given fields.","ref":"EctoFoundationDB.CLI.html#delete_fields!/4"},{"type":"function","title":"Arguments - EctoFoundationDB.CLI.delete_fields!/4","doc":"- `repo`: The repository to use.\n- `schema`: The schema to copy from.\n- `fields`: The list of fields to delete.\n- `options`: Options for the operation.","ref":"EctoFoundationDB.CLI.html#delete_fields!/4-arguments"},{"type":"function","title":"Requirements - EctoFoundationDB.CLI.delete_fields!/4","doc":"- Each field must not exist on the Schema.\n- Each field must not be used in any registered metadata (indexes).","ref":"EctoFoundationDB.CLI.html#delete_fields!/4-requirements"},{"type":"function","title":"Options - EctoFoundationDB.CLI.delete_fields!/4","doc":"- `:prefix`: If provided, only this tenant will be processed. If not provided, then all tenants concurrently.\n- `:max_rows`: Maximum number of items to process in a single transaction.\n\nAlso supports options from `run_concurrently_for_all_tenants!/3` and `EctoFoundationDB.Tenant.open/3`.","ref":"EctoFoundationDB.CLI.html#delete_fields!/4-options"},{"type":"function","title":"EctoFoundationDB.CLI.migrate!/2","doc":"Ensures all tenants in the database are migrated to the latest version.","ref":"EctoFoundationDB.CLI.html#migrate!/2"},{"type":"function","title":"Arguments - EctoFoundationDB.CLI.migrate!/2","doc":"- `repo`: The repository to use.\n- `options`: Options to pass to the migration function.","ref":"EctoFoundationDB.CLI.html#migrate!/2-arguments"},{"type":"function","title":"Options - EctoFoundationDB.CLI.migrate!/2","doc":"- `:prefix`: If provided, only this tenant will be migrated. If not provided, then all tenants concurrently.\n\nAlso supports options from `run_concurrently_for_all_tenants!/3` and `EctoFoundationDB.Tenant.open/3`.","ref":"EctoFoundationDB.CLI.html#migrate!/2-options"},{"type":"function","title":"EctoFoundationDB.CLI.run_concurrently_for_all_tenants!/3","doc":"Runs the provided 1-arity function for each tenant in the database concurrently.","ref":"EctoFoundationDB.CLI.html#run_concurrently_for_all_tenants!/3"},{"type":"function","title":"Arguments - EctoFoundationDB.CLI.run_concurrently_for_all_tenants!/3","doc":"- `repo`: The Ecto repository to use.\n- `caller_fun`: A 1-arity function; the argument is the tenant.\n- `options`: Optional keyword list of options.","ref":"EctoFoundationDB.CLI.html#run_concurrently_for_all_tenants!/3-arguments"},{"type":"function","title":"Options - EctoFoundationDB.CLI.run_concurrently_for_all_tenants!/3","doc":"- `:async_stream_options`: Options to be passed to `Task.async_stream/3`. Defaults to `ordered: false, max_concurrency: System.schedulers_online() * 2, timeout: :infinity`.","ref":"EctoFoundationDB.CLI.html#run_concurrently_for_all_tenants!/3-options"},{"type":"module","title":"EctoFoundationDB.Database","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Database.html"},{"type":"function","title":"EctoFoundationDB.Database.open/1","doc":"","ref":"EctoFoundationDB.Database.html#open/1"},{"type":"type","title":"EctoFoundationDB.Database.t/0","doc":"","ref":"EctoFoundationDB.Database.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Future","doc":"Opaque struct that represents an unresolved result set from some FoundationDB\nquery operation.\n\nIf you have received a Future from an EctoFDB Repo API, please consult the documentation\nfor that API for dealing with the Future. The functions here are intended to be for\ninternal use only.","ref":"EctoFoundationDB.Future.html"},{"type":"function","title":"EctoFoundationDB.Future.apply/2","doc":"","ref":"EctoFoundationDB.Future.html#apply/2"},{"type":"function","title":"EctoFoundationDB.Future.await_all/1","doc":"","ref":"EctoFoundationDB.Future.html#await_all/1"},{"type":"function","title":"EctoFoundationDB.Future.await_stream/1","doc":"","ref":"EctoFoundationDB.Future.html#await_stream/1"},{"type":"function","title":"EctoFoundationDB.Future.before_transactional/0","doc":"","ref":"EctoFoundationDB.Future.html#before_transactional/0"},{"type":"function","title":"EctoFoundationDB.Future.find_ready/2","doc":"","ref":"EctoFoundationDB.Future.html#find_ready/2"},{"type":"function","title":"EctoFoundationDB.Future.leaving_transactional/1","doc":"","ref":"EctoFoundationDB.Future.html#leaving_transactional/1"},{"type":"function","title":"EctoFoundationDB.Future.new/0","doc":"","ref":"EctoFoundationDB.Future.html#new/0"},{"type":"function","title":"EctoFoundationDB.Future.new_deferred/2","doc":"Creates a Future that will be resolved outside of the transaction in which it was created.\n\nUsed for:\n\n  - watch\n  - versionstamp","ref":"EctoFoundationDB.Future.html#new_deferred/2"},{"type":"function","title":"EctoFoundationDB.Future.ref/1","doc":"","ref":"EctoFoundationDB.Future.html#ref/1"},{"type":"function","title":"EctoFoundationDB.Future.result/1","doc":"","ref":"EctoFoundationDB.Future.html#result/1"},{"type":"function","title":"EctoFoundationDB.Future.set/4","doc":"","ref":"EctoFoundationDB.Future.html#set/4"},{"type":"function","title":"EctoFoundationDB.Future.set_result/2","doc":"","ref":"EctoFoundationDB.Future.html#set_result/2"},{"type":"function","title":"EctoFoundationDB.Future.token/0","doc":"","ref":"EctoFoundationDB.Future.html#token/0"},{"type":"module","title":"EctoFoundationDB.Index","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Index.html"},{"type":"type","title":"EctoFoundationDB.Index.t/0","doc":"","ref":"EctoFoundationDB.Index.html#t:t/0"},{"type":"behaviour","title":"EctoFoundationDB.Indexer","doc":"Implement this behaviour to create a custom index.\n\nEach Indexer has access to read, write, and **clear** any and all data in the database.\nA faulty implementation may lead to data loss or corruption.","ref":"EctoFoundationDB.Indexer.html"},{"type":"function","title":"EctoFoundationDB.Indexer.clear/5","doc":"","ref":"EctoFoundationDB.Indexer.html#clear/5"},{"type":"callback","title":"EctoFoundationDB.Indexer.clear/5","doc":"","ref":"EctoFoundationDB.Indexer.html#c:clear/5"},{"type":"function","title":"EctoFoundationDB.Indexer.create/6","doc":"","ref":"EctoFoundationDB.Indexer.html#create/6"},{"type":"callback","title":"EctoFoundationDB.Indexer.create/6","doc":"","ref":"EctoFoundationDB.Indexer.html#c:create/6"},{"type":"function","title":"EctoFoundationDB.Indexer.create_range/2","doc":"","ref":"EctoFoundationDB.Indexer.html#create_range/2"},{"type":"callback","title":"EctoFoundationDB.Indexer.create_range/2","doc":"","ref":"EctoFoundationDB.Indexer.html#c:create_range/2"},{"type":"function","title":"EctoFoundationDB.Indexer.drop_ranges/2","doc":"","ref":"EctoFoundationDB.Indexer.html#drop_ranges/2"},{"type":"callback","title":"EctoFoundationDB.Indexer.drop_ranges/2","doc":"","ref":"EctoFoundationDB.Indexer.html#c:drop_ranges/2"},{"type":"function","title":"EctoFoundationDB.Indexer.range/3","doc":"","ref":"EctoFoundationDB.Indexer.html#range/3"},{"type":"callback","title":"EctoFoundationDB.Indexer.range/3","doc":"","ref":"EctoFoundationDB.Indexer.html#c:range/3"},{"type":"function","title":"EctoFoundationDB.Indexer.set/5","doc":"","ref":"EctoFoundationDB.Indexer.html#set/5"},{"type":"callback","title":"EctoFoundationDB.Indexer.set/5","doc":"","ref":"EctoFoundationDB.Indexer.html#c:set/5"},{"type":"function","title":"EctoFoundationDB.Indexer.unpack/3","doc":"","ref":"EctoFoundationDB.Indexer.html#unpack/3"},{"type":"callback","title":"EctoFoundationDB.Indexer.unpack/3","doc":"","ref":"EctoFoundationDB.Indexer.html#c:unpack/3"},{"type":"function","title":"EctoFoundationDB.Indexer.update/6","doc":"","ref":"EctoFoundationDB.Indexer.html#update/6"},{"type":"callback","title":"EctoFoundationDB.Indexer.update/6","doc":"","ref":"EctoFoundationDB.Indexer.html#c:update/6"},{"type":"module","title":"EctoFoundationDB.Indexer.SchemaMetadata","doc":"This is Indexer keeps track of various actions for a Schema:\n\n- `inserts`: Incremented for each insert or upsert\n- `deletes`: Incremented for each delete\n- `collection`: Incremented for each insert, update, or delete\n- `updates`: Incremented for each update (via `Repo.update/*`)\n- `changes`: Incremented for each insert, upsert, delete, or update\n\nThese keys are useful for creating watches that will notify your application of those actions. For example, if you create a watch on the `inserts` key,\nyour application will be notified when a new record is inserted, and you can react however you like.\n\nSee it in action: [Sync Engine Part II - Collections](collection_syncing.html)","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.async_changes/1","doc":"Asynchronously get the `changes` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#async_changes/1"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.async_collection/1","doc":"Asynchronously get the `collection` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#async_collection/1"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.async_deletes/1","doc":"Asynchronously get the `deletes` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#async_deletes/1"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.async_inserts/1","doc":"Asynchronously get the `inserts` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#async_inserts/1"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.async_updates/1","doc":"Asynchronously get the `updates` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#async_updates/1"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.changes/2","doc":"Synchronous get on the `changes` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#changes/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.collection/2","doc":"Synchronous get on the `collection` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#collection/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.deletes/2","doc":"Synchronous get on the `deletes` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#deletes/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.inserts/2","doc":"Synchronous get on the `inserts` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#inserts/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.updates/2","doc":"Synchronous get on the `updates` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#updates/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.watch_changes/2","doc":"Create a watch on the `changes` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#watch_changes/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.watch_collection/2","doc":"Create a watch on the `collection` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#watch_collection/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.watch_deletes/2","doc":"Create a watch on the `deletes` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#watch_deletes/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.watch_inserts/2","doc":"Create a watch on the `inserts` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#watch_inserts/2"},{"type":"function","title":"EctoFoundationDB.Indexer.SchemaMetadata.watch_updates/2","doc":"Create a watch on the `updates` key.","ref":"EctoFoundationDB.Indexer.SchemaMetadata.html#watch_updates/2"},{"type":"module","title":"EctoFoundationDB.Layer","doc":"The Ecto FoundationDB Adapter implements a [Layer](https://apple.github.io/foundationdb/layer-concept.html)\non the underlying key-value store provided by FoundationDB. Via this layer, some\ncommon data access patterns are achieved. For those familiar with relational databases\nsuch as Postgres, these patterns will be familiar. However, there are many differences (for example SQL is\nnot supported), so this document seeks to describe the capabilities of the Ecto FoundationDB Layer in detail.","ref":"EctoFoundationDB.Layer.html"},{"type":"module","title":"Keyspace Design - EctoFoundationDB.Layer","doc":"All keys used by `:ecto_foundationdb` are encoded with [FoundationDB's Tuple Encoding](https://github.com/apple/foundationdb/blob/main/design/tuple.md).\n\nWith the default `:tenant_backend` of `EctoFoundationDB.Tenant.DirectoryTenant`, each tuple written by EctoFDB is\nprefixed with a short binary string as allocated by `:erlfdb_directory. This tenant prefixing is the most critical\nelement of the keyspace because it's the mechanism that guarantees tenants cannot cross their boundaries. **The rest of\nthis documentation is written assuming you're using the DirectoryTenant`.**\n\nThe first element of the tuple after the tenant prefix is a string prefix that is intended to keep the\n`:ecto_foundationdb` keyspace separate from other keys in the FoundationDB cluster.\n\nYour Schema data and Default indexes are stored with \"\\xFD\".\n\nThe data associated with schema migrations is stored with \"\\xFE\".\n\nThe rest of the tenant's keyspace is open for use by you, the application developer. For example, it is safe to write:\n\n```elixir\ndb = FoundationDB.open(MyApp.Repo)\ntenant = Tenant.open(MyApp.Repo, \"some-org\")\n\n# Multitenancy-Safe. The key is properly packed\n:erlfdb.set(db, Tenant.pack(tenant, {\"hello\"}), \"world\")\n\n# Multitenancy-Unsafe. The key is not packed into the tenant's keyspace\n:erlfdb.set(db, :erlfdb_tuple.pack({\"hello\"}), \"world\")\n:erlfdb.set(db, \"hello\", \"world\")\n```\n\nA value (the binary stored at each key) is either\n  * some other keys (in the case of Default indexes) or\n  * Erlang term data encoded with `:erlang.term_to_binary/1`","ref":"EctoFoundationDB.Layer.html#module-keyspace-design"},{"type":"module","title":"Primary Write and Read - EctoFoundationDB.Layer","doc":"Your Ecto Schema has a primary key field, which is usually a string or an integer. This primary\nkey uniquely identifies an object of your schema within the tenant in which it lives.\n\n```elixir\ndefmodule EctoFoundationDB.Schemas.User do\n  use Ecto.Schema\n  schema \"users\" do\n    field(:name, :string)\n    field(:department, :string)\n    timestamps()\n  end\nend\n```\n\nIn this example, a User has an `:id` and a `:name`. Also remember that the User is defined within\na tenant which provides a scope under which the User lives. For example, a typical tenant\nwould be the organization the User belongs to. Since the User is in this tenant, we do not need\nto provide an identifier for this organization on the User object itself.\n\n\"Primary Write\" refers to the insertion of the User struct into the FoundationDB key-value store\nunder a single key that uniquely identifies the User. This key includes the `:id` value.\n\nYour struct data is stored as a `Keyword` encoded with `:erlang.term_to_binary/1`.\n\nNote: The Primary Write can be skipped by providing the `write_primary: false` option on the `@schema_context`.\nSee below for more.","ref":"EctoFoundationDB.Layer.html#module-primary-write-and-read"},{"type":"module","title":"Default Indexes - EctoFoundationDB.Layer","doc":"When a Default index is created via a migration, the Ecto FoundationDB Adapter writes a set of\nkeys and values to facilitate lookups based on the indexed field.","ref":"EctoFoundationDB.Layer.html#module-default-indexes"},{"type":"module","title":"Advanced Options: `write_primary: false`, `mapped?: false` - EctoFoundationDB.Layer","doc":"If you choose to use `write_primary: false` on your schema, this skips the Primary Write. The consequence of this are as follows:\n\n 1. You'll want to make sure your index is created with an option `mapped?: false`. This ensures that the\n    struct data is written to the index keyspace.\n 2. Your index queries will now have performance characteristics similar to a primary key query. That is, you'll\n    be able to retrieve the struct data with a single `erlfdb:get_range/3`.\n 3. The data can **only** be managed by providing a query on the index. You will not be able to access the data\n    via the primary key.\n 4. If `write_primary: false`, then only one index can be created.","ref":"EctoFoundationDB.Layer.html#module-advanced-options-write_primary-false-mapped-false"},{"type":"behaviour","title":"EctoFoundationDB.Migrator","doc":"Implement this behaviour to define migrations for `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Migrator.html"},{"type":"callback","title":"EctoFoundationDB.Migrator.migrations/0","doc":"","ref":"EctoFoundationDB.Migrator.html#c:migrations/0"},{"type":"function","title":"EctoFoundationDB.Migrator.up/3","doc":"","ref":"EctoFoundationDB.Migrator.html#up/3"},{"type":"module","title":"EctoFoundationDB.Options","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.Options.html"},{"type":"function","title":"EctoFoundationDB.Options.get/2","doc":"","ref":"EctoFoundationDB.Options.html#get/2"},{"type":"type","title":"EctoFoundationDB.Options.option/0","doc":"","ref":"EctoFoundationDB.Options.html#t:option/0"},{"type":"type","title":"EctoFoundationDB.Options.t/0","doc":"","ref":"EctoFoundationDB.Options.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.QueryPlan","doc":"See `Ecto.Adapters.FoundationDB`","ref":"EctoFoundationDB.QueryPlan.html"},{"type":"function","title":"EctoFoundationDB.QueryPlan.all_range/7","doc":"","ref":"EctoFoundationDB.QueryPlan.html#all_range/7"},{"type":"function","title":"EctoFoundationDB.QueryPlan.get/7","doc":"","ref":"EctoFoundationDB.QueryPlan.html#get/7"},{"type":"function","title":"EctoFoundationDB.QueryPlan.get_op/3","doc":"","ref":"EctoFoundationDB.QueryPlan.html#get_op/3"},{"type":"function","title":"EctoFoundationDB.QueryPlan.get_pinned_param/2","doc":"","ref":"EctoFoundationDB.QueryPlan.html#get_pinned_param/2"},{"type":"function","title":"EctoFoundationDB.QueryPlan.walk_ast/4","doc":"","ref":"EctoFoundationDB.QueryPlan.html#walk_ast/4"},{"type":"type","title":"EctoFoundationDB.QueryPlan.t/0","doc":"","ref":"EctoFoundationDB.QueryPlan.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Sandbox","doc":"A module for managing a sandbox FoundationDB cluster. This allows a developer to create\na space under which it should be safe to write tests.\n\nWhen using this module, it creates a directory in the project root called `.erlfdb`. It\nis safe to delete this directory when you no longer need it (e.g. after test execution)\n\nSee [Testing with EctoFoundationDB](testing.html) for more.","ref":"EctoFoundationDB.Sandbox.html"},{"type":"function","title":"EctoFoundationDB.Sandbox.checkin/2","doc":"","ref":"EctoFoundationDB.Sandbox.html#checkin/2"},{"type":"function","title":"EctoFoundationDB.Sandbox.checkout/3","doc":"","ref":"EctoFoundationDB.Sandbox.html#checkout/3"},{"type":"function","title":"EctoFoundationDB.Sandbox.open_db/1","doc":"","ref":"EctoFoundationDB.Sandbox.html#open_db/1"},{"type":"module","title":"EctoFoundationDB.Tenant","doc":"This module allows the application to create, open, and delete\ntenants within the FoundationDB database. All transactions require a tenant,\nso any application that uses the Ecto FoundationDB Adapter must use this module.\n\nEctoFoundationDB supports 2 different backends for the Tenant implementation:\n\n1. `EctoFoundationDB.Tenant.DirectoryTenant`: The default. Tenants are managed via\n   the `:erlfdb_directory` layer. Each tenant is assigned a byte prefix. EctoFDB\n   puts that prefix as the first element of every tuple key. If your application wishes\n   to write keys directly to the database, it must also respect this prefix.\n2. `EctoFoundationDB.Tenant.ManagedTenant`: Uses the experimental [FoundationDB Tenants](https://apple.github.io/foundationdb/tenants.html).\n   This requires `tenant_mode=required_experimental` or `tenant_mode=optional_experimental`\n   to be enabled on your FDB system. EctoFDB does not need to prefix keys because it happens\n   automatically at the FDB transaction level. If your application wishes to write keys\n   directory to the database, there is no prefix to worry about. **Use caution:** FDB Tenants\n   are not yet tested with the same rigor as the rest of FDB.\n\nIf you wish to use the experimental ManagedTenant, add this option to your Repo config:\n\n```elixir\nconfig :my_app, MyApp.Repo,\n  tenant_backend: EctoFoundationDB.Tenant.ManagedTenant\n```\n\nEctoFDB does not support switching between tenant backends on a database. If you want to switch backend,\nyou must use a new empty database.","ref":"EctoFoundationDB.Tenant.html"},{"type":"function","title":"EctoFoundationDB.Tenant.clear/2","doc":"Clear all data for the given tenant. This cannot be undone.","ref":"EctoFoundationDB.Tenant.html#clear/2"},{"type":"function","title":"EctoFoundationDB.Tenant.clear_delete!/2","doc":"Clears data in a tenant and then deletes it. If the tenant doesn't exist, no-op.","ref":"EctoFoundationDB.Tenant.html#clear_delete!/2"},{"type":"function","title":"EctoFoundationDB.Tenant.create/2","doc":"Create a tenant in the database.","ref":"EctoFoundationDB.Tenant.html#create/2"},{"type":"function","title":"EctoFoundationDB.Tenant.delete/2","doc":"Deletes a tenant from the database permanently. The tenant must\nhave no data.","ref":"EctoFoundationDB.Tenant.html#delete/2"},{"type":"function","title":"EctoFoundationDB.Tenant.exists?/2","doc":"Returns true if the tenant already exists in the database.","ref":"EctoFoundationDB.Tenant.html#exists?/2"},{"type":"function","title":"EctoFoundationDB.Tenant.extend_tuple/2","doc":"","ref":"EctoFoundationDB.Tenant.html#extend_tuple/2"},{"type":"function","title":"EctoFoundationDB.Tenant.list/1","doc":"List all tenants in the database. Could be expensive.","ref":"EctoFoundationDB.Tenant.html#list/1"},{"type":"function","title":"EctoFoundationDB.Tenant.open/3","doc":"Open a tenant with a repo. With the result returned by this function, the caller can\ndo database operations on the tenant's portion of the key-value store.\n\nThe tenant must already exist.\n\nWhen opening tenants with a repo, all migrations are automatically performed. This\ncan cause open/2 to take a significant amount of time. Tenants can be kept open\nindefinitely, with any number of database transactions issued upon them.","ref":"EctoFoundationDB.Tenant.html#open/3"},{"type":"function","title":"EctoFoundationDB.Tenant.open!/3","doc":"Open a tenant. With the result returned by this function, the caller can\ndo database operations on the tenant's portion of the key-value store.\n\nIf the tenant does not exist, it is created.\n\nWhen opening tenants with a repo, all migrations are automatically performed. This\ncan cause open/2 to take a significant amount of time. Tenants can be kept open\nindefinitely, with any number of database transactions issued upon them.","ref":"EctoFoundationDB.Tenant.html#open!/3"},{"type":"function","title":"EctoFoundationDB.Tenant.open_empty!/3","doc":"Helper function to ensure the given tenant exists and then clear\nit of all data, and finally return an open handle. Useful in test code,\nbut in production, this would be dangerous.","ref":"EctoFoundationDB.Tenant.html#open_empty!/3"},{"type":"function","title":"EctoFoundationDB.Tenant.pack/2","doc":"Packs an Elixir tuple into an FDB-encoded Tuple.","ref":"EctoFoundationDB.Tenant.html#pack/2"},{"type":"function","title":"Keyspace Design - EctoFoundationDB.Tenant.pack/2","doc":"We always pack into a **non-prefixed tuple key**. In other words, EctoFDB\nDirectoryTenant keys use the subspace prefix as the first tuple element\ninstead of binary key prefix. As such, our keys are not compliant with\nother Directory Layer implementations. However, we've made this choice so that\nwe can continue to use GetMappedRange functionality. (Indeed, the GetMappedRange mapper spec\nhas reserved syntax for stripping out non-tuple prefixes, but it's not yet implemented.)\nNote: ManagedTenant keys do not use directories/subspaces. The underlying binary\nprefix that a ManagedTenant uses internally still allows use of GetMappedRange.\n\nManagedTenants and GetMappedRange are both experimental features. The safest choice is\nto use neither, but that would forfeit the GetMappedRange optimization. We choose\nto accept the risk of GetMappedRange, which can easily be replaced with a different\nclient implementation, and suggest against using ManagedTenant because it puts\nyour data at risk with its unsupported[[0](https://github.com/apple/foundationdb/issues/11292)][[1](https://github.com/apple/foundationdb/issues/11382)]\nkeyspace.","ref":"EctoFoundationDB.Tenant.html#pack/2-keyspace-design"},{"type":"function","title":"EctoFoundationDB.Tenant.pack_vs/2","doc":"Packs an Elixir tuple having an incomplete versionstamp into an FDB-encoded Tuple.\n\nCaller should proceed to use `:erlfdb.set_versionstamped_key` or `:erlfdb.set_versionstamped_value`\nas needed.","ref":"EctoFoundationDB.Tenant.html#pack_vs/2"},{"type":"function","title":"EctoFoundationDB.Tenant.primary_codec/3","doc":"","ref":"EctoFoundationDB.Tenant.html#primary_codec/3"},{"type":"function","title":"EctoFoundationDB.Tenant.range/2","doc":"","ref":"EctoFoundationDB.Tenant.html#range/2"},{"type":"function","title":"EctoFoundationDB.Tenant.set_metadata_cache/2","doc":"","ref":"EctoFoundationDB.Tenant.html#set_metadata_cache/2"},{"type":"function","title":"EctoFoundationDB.Tenant.txobj/1","doc":"","ref":"EctoFoundationDB.Tenant.html#txobj/1"},{"type":"function","title":"EctoFoundationDB.Tenant.unpack/2","doc":"","ref":"EctoFoundationDB.Tenant.html#unpack/2"},{"type":"type","title":"EctoFoundationDB.Tenant.id/0","doc":"","ref":"EctoFoundationDB.Tenant.html#t:id/0"},{"type":"type","title":"EctoFoundationDB.Tenant.prefix/0","doc":"","ref":"EctoFoundationDB.Tenant.html#t:prefix/0"},{"type":"type","title":"EctoFoundationDB.Tenant.t/0","doc":"","ref":"EctoFoundationDB.Tenant.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Tenant.DirectoryTenant","doc":"The default backend for EctoFDB multitenancy.\n\nIt uses the Directory Layer defined by `:erlfdb_directory` to\npartition tenant keyspaces from each other. There is no guarantee\nprovided by FDB at the transaction-level, so care must be taken to\nalways pack keys appropriately. Standard use of Ecto.Repo functions\nis guaranteed to respect tenant boundaries.","ref":"EctoFoundationDB.Tenant.DirectoryTenant.html"},{"type":"type","title":"EctoFoundationDB.Tenant.DirectoryTenant.t/0","doc":"","ref":"EctoFoundationDB.Tenant.DirectoryTenant.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Tenant.ManagedTenant","doc":"An experimental backend for EctoFDB multitenancy. It uses FDB's Tenants\nto guarantee that a given transaction cannot access keys from another tenant.\n\nTo use ManagedTenant, your database must be configured with\n\n```shell\nfdbcli --exec 'configure tenant_mode=required_experimental'\n# or\nfdbcli --exec 'configure tenant_mode=optional_experimental'\n```\n\nFDB's tenants are still an experimental feature, so EctoFDB's ManagedTenant must\nalso be considered experimental. For now, we recommend using `EctoFoundationDB.Tenant.DirectoryTenant`,\nwhich is the default.","ref":"EctoFoundationDB.Tenant.ManagedTenant.html"},{"type":"type","title":"EctoFoundationDB.Tenant.ManagedTenant.t/0","doc":"","ref":"EctoFoundationDB.Tenant.ManagedTenant.html#t:t/0"},{"type":"module","title":"EctoFoundationDB.Versionstamp","doc":"Versionstamping is a feature that allows you to create a unique identifier for a record\nthat is guaranteed to be unique across all records in the database.\n\nPlease refer to the documentation for `Repo.async_insert_all!/3`.","ref":"EctoFoundationDB.Versionstamp.html"},{"type":"function","title":"EctoFoundationDB.Versionstamp.embed_as/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#embed_as/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.equal?/2","doc":"","ref":"EctoFoundationDB.Versionstamp.html#equal?/2"},{"type":"function","title":"EctoFoundationDB.Versionstamp.from_binary/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#from_binary/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.from_integer/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#from_integer/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.get/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#get/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.incomplete/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#incomplete/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.incomplete?/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#incomplete?/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.next/0","doc":"","ref":"EctoFoundationDB.Versionstamp.html#next/0"},{"type":"function","title":"EctoFoundationDB.Versionstamp.next/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#next/1"},{"type":"function","title":"EctoFoundationDB.Versionstamp.resolve/2","doc":"","ref":"EctoFoundationDB.Versionstamp.html#resolve/2"},{"type":"function","title":"EctoFoundationDB.Versionstamp.to_integer/1","doc":"","ref":"EctoFoundationDB.Versionstamp.html#to_integer/1"},{"type":"exception","title":"EctoFoundationDB.Exception.IncorrectTenancy","doc":"This exception is raised when there is a conflict in the tenant used.","ref":"EctoFoundationDB.Exception.IncorrectTenancy.html"},{"type":"exception","title":"EctoFoundationDB.Exception.Unsupported","doc":"This exception is raised when the application uses an Ecto feature that is not\nsupported by the FoundationDB Adapter.","ref":"EctoFoundationDB.Exception.Unsupported.html"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"v0.5.0 (2025-07-08) - Changelog","doc":"","ref":"changelog.html#v0-5-0-2025-07-08"},{"type":"extras","title":"Enhancements - Changelog","doc":"* `EctoFoundationDB.CLI` [[doc](operators_manual.md)]: Functions for an operator to use to safely manage\ndata migrations that cannot be done automatically.\n* Index Metadata [[doc](metadata.html)] now makes use of FoundationDB's `\\xff/metadataVersion` key,\nwhich allows the client to cache metadata and maintain transactional isolation\nwithout having to wait on any keys. Also, the cache is shared across all open tenants of the same id on a given node.\n* `EctoFoundationDB.Versionstamp` [[doc](Ecto.Adapters.FoundationDB.html#module-versionstamps-autoincrement)]: Added the ability to insert objects with a monotonically increasing integer id, via FoundationDB's versionstamp.\n* `SchemaMetadata` [[doc](Ecto.Adapters.FoundationDB.html#module-schema-metadata)]: This is a new built-in Indexer that allows your app to watch and sync collection of objects in a tenant.\n* Added ability to drop an index.","ref":"changelog.html#enhancements"},{"type":"extras","title":"Breaking changes - Changelog","doc":"* Key construction has changed, making databases created on <= 0.4.x incompatible with >=0.5. Specifically, a binary, atom, or\n  number primary key is now encoded in the FDB key with the Tuple layer. All other types remain encoded with term_to_binary.\n  If you need help upgrading your database, please put in a GitHub Issue. We strive for a stable v1.0.","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Fixed a bug where index creation was failing for multikey objects\n* (#57) Fixed a bug where index management was failing while a new index was being created","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"Deprecations - Changelog","doc":"* Ecto has deprecated `Repo.transaction` in favor of `Repo.transact`. Since this decision doesn't align with FoundationDB's view of\ntransactions, we have chosen to deprecate `Repo.transaction` in favor of `Repo.transactional`. This terminology better aligns with\n`:erlfdb` and provides a distinction from RDBMS transactions, and allows us to avoid future deprecations.","ref":"changelog.html#deprecations"},{"type":"extras","title":"New documentation - Changelog","doc":"* [Guide for Operators](operators_manual.html): Describes how to use the `EctoFoundationDB.CLI` functions to rename a field while guaraneeting that all\nconcurrent queries in your distributed application are successful.\n* [Metadata Design](metadata.html): Describes how index metadata is managed and cached in EctoFDB.\n* [Sync Engine Part I - Single Object](watches.livemd): Revamped Livebook that demonstrates how to create a Sync Engine for a single object (for syncing reads)\n* [Sync Engine Part II - Collections](collection_syncing.livemd): New Livebook that demonstrates a Sync Engine for a collection of objects in a tenant (still for reads)","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.4.0 (2025-01-16) - Changelog","doc":"","ref":"changelog.html#v0-4-0-2025-01-16"},{"type":"extras","title":"Enhancements - Changelog","doc":"* [Large Structs](Ecto.Adapters.FoundationDB.html#module-advanced-options): If your struct encodes to a size larger than 100,000 Bytes, it will now be split across several FDB key-values automatically.\n  Previously, EctoFDB did not attempt to detect this and allowed FoundationDB to throw error code 2103: \"value_too_large - Value length exceeds limit\".\n  See documentation for configuration options.\n* `EctoFoundationDB.Sandbox` now uses `m:erlfdb_sandbox`. Sandbox directory name is now `.erlfdb_sandbox`. Directories named `.erlfdb` should be removed.","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Upgrade erlfdb to v0.2.2","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.3.1 (2024-10-24) - Changelog","doc":"","ref":"changelog.html#v0-3-1-2024-10-24"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Fixed consistency issue with index updates. Previously, the old index key was still queryable.\n* Fixed write amplification issue when updating struct's non-indexed fields.","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"New Documentation - Changelog","doc":"* Added `fdb_api_counting_text.exs` which tests and documents the `:erlfdb` operations that our Layer is expected to make.","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.3.0 (2024-10-20) - Changelog","doc":"### \\*\\* Major breaking changes \\*\\*\n\nDatabases that have been created using a prior version of EctoFoundationDB will be broken on\nEctoFDB v0.3 and above. Please start a new database with EctoFDB v0.3. If you currently have\na database on v0.2 or earlier, please [submit an issue](https://github.com/foundationdb-beam/ecto_foundationdb/issues)\nto discuss the upgrade path.","ref":"changelog.html#v0-3-0-2024-10-20"},{"type":"extras","title":"Enhancements - Changelog","doc":"* [Watches](Ecto.Adapters.FoundationDB.html#module-watches): Support for FDB Watches, which is like a database-driven PubSub on an Ecto struct.\n* [Directory Tenants](EctoFoundationDB.Tenant.html): A new default backend for Multitenancy that is production-ready. Managed tenants have been moved to \"Experimental\" status.\n* `@schema_context usetenant: true` is no longer required.\n* The `:open_db` option now defines a 1-arity function that accepts the Repo module.","ref":"changelog.html#enhancements"},{"type":"extras","title":"New Documentation - Changelog","doc":"* [Livebook | Watches in LiveView](watches.livemd)","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.2.1 (2024-09-23) - Changelog","doc":"","ref":"changelog.html#v0-2-1-2024-09-23"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Upgrade erlfdb to 0.2.1, allowing Livebook Desktop to discover fdbcli location in /usr/local/bin","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.2.0 (2024-09-21) - Changelog","doc":"","ref":"changelog.html#v0-2-0-2024-09-21"},{"type":"extras","title":"Bug fixes - Changelog","doc":"","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"Enhancements - Changelog","doc":"* [Upserts](Ecto.Adapters.FoundationDB.html#module-upserts): Support for Ecto options `:on_conflict` and `:conflict_target`\n  * [Pipelining](Ecto.Adapters.FoundationDB.html#module-pipelining): New Repo functions for async/await within a transaction.","ref":"changelog.html#enhancements"},{"type":"extras","title":"New Documentation - Changelog","doc":"* [testing.md](testing.html): Document to describe how to set up a Sandbox\n  * [CHANGELOG.md](changelog.html): This file!","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.1.2 (2024-08-31) - Changelog","doc":"","ref":"changelog.html#v0-1-2-2024-08-31"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Upgrade erlfdb","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.1.1 (2024-08-25) - Changelog","doc":"","ref":"changelog.html#v0-1-1-2024-08-25"},{"type":"extras","title":"Enhancements - Changelog","doc":"* Upgrade to Ecto 3.12","ref":"changelog.html#enhancements"},{"type":"extras","title":"New documentation - Changelog","doc":"* [Livebook | Getting Started](introduction.livemd): How to get started with EctoFoundationDB.","ref":"changelog.html#new-documentation"},{"type":"extras","title":"v0.1.0 (2024-04-07) - Changelog","doc":"","ref":"changelog.html#v0-1-0-2024-04-07"},{"type":"extras","title":"Features - Changelog","doc":"* Multitenancy\n* Basic CRUD operations\n* Indexes","ref":"changelog.html#features"},{"type":"extras","title":"Introduction","doc":"# Introduction\n\n```elixir\nMix.install([\n  {:ecto_foundationdb, \"~> 0.3\"}\n])\n```","ref":"introduction.html"},{"type":"extras","title":"Setup - Introduction","doc":"Hello! This guide simulates what your experience might be when developing an application with EctoFoundationDB. Specifically, it focuses on the mechanism that EctoFoundationDB uses to create and manage indexes.\n\nIt assumes the reader is familiar with general Ecto features.\n\nBefore we get started, a couple of important points about executing these commands on your system.\n\n> If you received an error on the `Mix.install` setup, please make sure you have both `foundationdb-server` and `foundationdb-clients` packages installed on your system. Also, ensure that your Livebook PATH environment variable includes the directory containing the `fdbcli` binary.\n\n> This LiveBook expects your system to have a running instance of FoundationDB, and it writes and deletes data from it. If your system's `/etc/foundationdb/fdb.cluster` is pointing to a real database, do not execute these commands!\n\nWith that out of the way, we'll start off with creating your Repo module.\n\n```elixir\ndefmodule MyApp.Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n\n  @impl true\n  def migrations() do\n    [\n      # {0, IndexesMigration}\n    ]\n  end\nend\n```\n\nNotice that the line with `IndexesMigration` is commented out. We'll come back to this later.","ref":"introduction.html#setup"},{"type":"extras","title":"Developing your app - Introduction","doc":"This next step simulates your app's startup. Normally, you would have a project defining `:my_app` and the Repo would be included in your supervision tree. In this Guide, we're starting the Repo as an isolated resource.\n\n```elixir\n{:ok, _} = Ecto.Adapters.FoundationDB.ensure_all_started(MyApp.Repo.config(), :temporary)\nMyApp.Repo.start_link(log: false)\n```\n\nNext, we define an `Ecto.Schema` for events that are coming from a temperature sensor. This is a pretty standard Schema module.\n\n```elixir\ndefmodule TemperatureEvent do\n  use Ecto.Schema\n\n  @primary_key {:id, :binary_id, autogenerate: true}\n\n  schema \"temperature_events\" do\n    field(:recorded_at, :naive_datetime_usec)\n    field(:kelvin, :float)\n    field(:site, :string)\n    timestamps()\n  end\nend\n```\n\nWe're going to create a module that will help us insert some `TemperatureEvents`.\n\n```elixir\ndefmodule Sensor do\n  alias Ecto.Adapters.FoundationDB\n\n  def record(n, tenant) do\n    MyApp.Repo.transactional(tenant,\n      fn ->\n        for _ <- 1..n, do: record(nil)\n      end)\n  end\n\n  def record(tenant) do\n    %TemperatureEvent{\n      site: \"surface\",\n      kelvin: 373.15 + :rand.normal(0, 5),\n      recorded_at: NaiveDateTime.utc_now()\n    }\n    |> FoundationDB.usetenant(tenant)\n    |> MyApp.Repo.insert!()\n  end\nend\n```\n\nNow, we create and open a new Tenant to store our `TemperatureEvents`.\n\n```elixir\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(MyApp.Repo, \"experiment-42c\")\n```\n\nWe're ready to record an event from our temperature sensor! Feel free to Reevaluate this block several times. You'll record 4 new events each time.\n\n```elixir\nfor _ <- 1..4, do: Sensor.record(tenant)\n```\n\nWe can list all the events from the Tenant. This uses a single FoundationDB Transaction.\n\n```elixir\nMyApp.Repo.all(TemperatureEvent, prefix: tenant)\n```\n\nIf there's a large number of events, you can stream them instead of reading them all at once. This uses multiple FoundationDB Transactions.\n\n```elixir\nMyApp.Repo.stream(TemperatureEvent, prefix: tenant)\n|> Enum.to_list()\n|> length()\n```\n\nNext, we'd like to read all events from `\"surface\"`. If you're executing this LiveBook in order, you'll receive an exception on this step.\n\n```elixir\nimport Ecto.Query\n\nquery = from(e in TemperatureEvent, where: e.site == ^\"surface\")\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nDid you get an exception? If so, scroll back up to the `defmodule MyApp.Repo` block in the Setup section, un-comment the line with `IndexesMigration`, and Reevaluate that block. Then come back and continue from here. You don't need to Reevaluate other blocks above this text.\n\n👋\n\nWelcome back! You've instructed the Repo to load a migration next time we open a Tenant. But we still need to define that Migration. The block below defines two indexes.\n\n```elixir\ndefmodule IndexesMigration do\n  use EctoFoundationDB.Migration\n\n  @impl true\n  def change() do\n    [\n      create(index(TemperatureEvent, [:site])),\n      create(index(TemperatureEvent, [:recorded_at]))\n    ]\n  end\nend\n```\n\nNow, we re-open the Tenant. **Something very important happens here.**\n\nThis block simulates you restarting your app, and your client reconnecting. We'll just simply call `open!/2` again.\n\n```elixir\ntenant = Tenant.open!(MyApp.Repo, \"experiment-42c\")\n```\n\nGreat! If you made it to this step, then the Migration has executed automatically, and the indexes are ready to be used.","ref":"introduction.html#developing-your-app"},{"type":"extras","title":"Querying your data - Introduction","doc":"This next block has the same query as the one that threw an exception earlier. This time, you should retrieve the expected events.\n\n```elixir\nimport Ecto.Query\nquery = from(e in TemperatureEvent, where: e.site == ^\"surface\")\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nWe can also use the timestamp index that we created in a new query.\n\n```elixir\nnow = NaiveDateTime.utc_now()\npast = NaiveDateTime.add(now, -1200, :second)\n\nquery =\n  from(e in TemperatureEvent,\n    where: e.recorded_at >= ^past and e.recorded_at < ^now\n  )\n\nMyApp.Repo.all(query, prefix: tenant)\n```\n\nFinally, just for fun, let's insert 10,000 `TemperatureEvent`s!\n\n```elixir\nnum = 10000\nbatch = 100\n\n{t, :ok} =\n  :timer.tc(fn ->\n    Stream.duplicate(batch, div(num, batch))\n    |> Task.async_stream(\n      Sensor,\n      :record,\n      [tenant],\n      max_concurrency: System.schedulers_online() * 8,\n      ordered: false,\n      timeout: 30000\n    )\n    |> Stream.run()\n  end)\n\nIO.puts(\"Done in #{t / 1000} msec\")\n```","ref":"introduction.html#querying-your-data"},{"type":"extras","title":"Cleaning up - Introduction","doc":"And if you'd like to tidy up, you can easily delete all the data.\n\n```elixir\n# Note: destructive!\nMyApp.Repo.delete_all(TemperatureEvent, prefix: tenant)\n```\n\n```elixir\n# Note: destructive!\nTenant.clear_delete!(MyApp.Repo, \"experiment-42c\")\n```","ref":"introduction.html#cleaning-up"},{"type":"extras","title":"Sync Engine Part I - Single Object","doc":"# Sync Engine Part I - Single Object\n\n```elixir\nMix.install([\n  {:phoenix_playground, \"~> 0.1.0\"},\n  {:ecto_foundationdb, github: \"foundationdb-beam/ecto_foundationdb\"}\n])\n```","ref":"watches.html"},{"type":"extras","title":"Intro - Sync Engine Part I - Single Object","doc":"Using [FoundationDB Watches](https://hexdocs.pm/ecto_foundationdb/Ecto.Adapters.FoundationDB.html#module-watches), we can set up a mini read-path Sync Engine. It will automatically propagate new data to all mounted LiveViews with push-messaging delivered from the database directly to the LiveView process.\n\nIn this demo, we'll update our \"favorite quote\" in the database and discuss the mechanism used to update the front-end with new data.","ref":"watches.html#intro"},{"type":"extras","title":"Setup Ecto - Sync Engine Part I - Single Object","doc":"First, we set up Ecto, defining a Quote schema, and starting the Repo. For this to work, you must have `foundationdb-server` installed. Refer to the [EctoFoundationDB documentation](https://hexdocs.pm/ecto_foundationdb/Ecto.Adapters.FoundationDB.html#module-installation) for installation help.\n\n```elixir\ndefmodule Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n\n  @impl true\n  def migrations(), do: []\nend\n\ndefmodule Quote do\n  use Ecto.Schema\n\n  alias Ecto.Changeset\n\n  @primary_key {:id, :binary_id, autogenerate: true}\n\n  schema \"quotes\" do\n    field(:author, :string)\n    field(:content, :string)\n    field(:likes, :integer, default: 0)\n    field(:as_of, :naive_datetime)\n    timestamps()\n  end\n\n  def changeset(quote, params \\\\ %{}) do\n    quote\n    |> Changeset.cast(params, [:author, :content, :likes, :as_of])\n  end\n\n  def like!(tenant, id) do\n      Repo.transactional(tenant, fn ->\n        quote = Repo.get!(Quote, id)\n\n        quote\n        |> changeset(%{likes: quote.likes+1})\n        |> Repo.update!()\n      end)\n  end\nend\n\n# ------------------------------------------------------------------\n# This section contains some set-up within Livebook that's typically\n# handled instead in your application startup and config.\nApplication.put_env(:my_app, Repo,\n  open_db: &EctoFoundationDB.Sandbox.open_db/1,\n  storage_id: EctoFoundationDB.Sandbox\n)\n\n{:ok, _} = Ecto.Adapters.FoundationDB.ensure_all_started(Repo.config(), :temporary)\nRepo.start_link(log: false)\n# ------------------------------------------------------------------\n\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview\")\n\n```\n\nNext, we create a Quote in the DB. We'll focus on how this specific Quote is rendered throughout the rest of the Livebook.\n\n```elixir\nRepo.insert!(\n  %Quote{\n    id: \"my-favorite-quote\",\n    author: \"Philippe Verdoux\",\n    content: \"\"\"\n      Enlightenment leads to benightedness; Science entails nescience.\n      \"\"\",\n    as_of: NaiveDateTime.utc_now(:second)\n  }, prefix: tenant, on_conflict: :replace_all)\n```","ref":"watches.html#setup-ecto"},{"type":"extras","title":"Setup LiveView - Sync Engine Part I - Single Object","doc":"We're using Phoenix Playground to create a sample LiveView. The single code block below packs a bunch of things together. Here are the key takeways:\n\n1. In `sync_one!/4`, we're initializing our mini sync engine.\n\n   We do a standard get on the Quote with id `\"my-favorite-quote\"` that we've inserted above.\n\n   In the same transaction, we're creating a watch with `label: :quote`.\n\n   Finally, we store the `:quote` and the list of `:futures`.\n\n2. `handle_ready/2` is a LiveView hook that receives the `:ready` message from the watch future, which is delivered directly from the database layer. We use `Repo.assign_ready/3` to update the LiveView assigns map. Whenever `\"my-favorite-quote\"` changes in the DB, this function will be called automatically. Also, we provide `watch?: true` to continue listening for updates.\n\n3. `handle_event/3` receives the `\"like\"` event, and uses an FDB transaction to add 1 to the count of likes. Notice we choose not to update the assigns, for demonstration purposes.\n\n4. `sync_one!/4` and `handle_ready/2` both contain nothing specific to the Quote schema. This is a general purpose approach that can be re-used for any other schema types.\n\n```elixir\ndefmodule DemoLive do\n  use Phoenix.LiveView\n\n  def mount(_params, _session, socket) do\n\n    tenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview\")\n\n    {:ok, socket\n      |> assign(tenant: tenant)\n      |> sync_one!(Quote, :quote, \"my-favorite-quote\")\n    }\n  end\n\n  # 1. Initialize our mini Sync Engine\n  defp sync_one!(socket, schema, label, id) do\n    {struct, futures} = Repo.transactional(socket.assigns.tenant,\n      fn ->\n        struct = Repo.get!(schema, id)\n        future = Repo.watch(struct, label: label)\n        {struct, [future]}\n      end)\n\n    socket\n    |> assign(label, struct)\n    |> assign(futures: futures)\n    |> attach_hook({:assign_ready, schema, label, id}, :handle_info, &handle_ready/2)\n  end\n\n  # 2. LiveView server hook that receives a message from the database layer\n  #    and updates the assigns\n  defp handle_ready({ref, :ready}, socket) do\n\n    %{assigns: assigns} = socket\n\n    {new_assigns, futures} = Repo.assign_ready(\n        assigns.futures,\n        [ref],\n        watch?: true,\n        prefix: assigns.tenant\n      )\n    {:halt, socket\n            |> assign(new_assigns)\n            |> assign(futures: futures)}\n  end\n\n  def render(assigns) do\n    ~H\"\"\"\n     \n       My Favorite Quote \n       as of <%= @quote.as_of %> UTC \n       \n         <%= @quote.content %> \n          - <%= @quote.author %>  \n\n         \n           Likes:  \n           <%= @quote.likes %> \n           👍 \n         \n       \n     \n\n     \n      h1 { text-align: center; }\n      h3 { text-align: right; }\n      h6 { text-align: right; }\n      p { text-align: center; }\n\n      .content {\n        max-width: 500px;\n        margin: auto;\n      }\n     \n    \"\"\"\n  end\n\n  # 3. Receive the \"like\" event and update the DB\n  def handle_event(\"like\", _params, socket) do\n\n    Quote.like!(socket.assigns.tenant, socket.assigns.quote.id)\n\n    {:noreply, socket}\n  end\n\nend\n\nPhoenixPlayground.start(live: DemoLive)\n```\n\nWith the block above evaluated, you can now open a web browser to `http://localhost:4000`. We suggest you keep this open side-by-side with this Livebook if possible, so that you can watch the updates in real-time. You can even open several tabs to ensure that all tabs are indeed in sync.\n\nYou can click the Like button and observe that the value changes. Notice that we are not pushing this value to the assigns explicitly. It's being handled automatically in our mini sync engine in `handle_info`.","ref":"watches.html#setup-liveview"},{"type":"extras","title":"Updating My Favorite Quote - Sync Engine Part I - Single Object","doc":"Your favorite quote might change from moment to moment. We'll write to the database with your current favorite quote. Because of the FDB watch, the LiveView will always render the most up-to-date content.\n\n```elixir\ntenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview\")\n\ndefmodule ChangeFavoriteQuote do\n  @id \"my-favorite-quote\"\n\n  def to(tenant, author, content) do\n    Repo.transactional(tenant, fn ->\n      Repo.get(Quote, @id)\n      |> Quote.changeset(%{\n        author: author,\n        content: content,\n        as_of: NaiveDateTime.utc_now(:second),\n        likes: 0\n      })\n      |> Repo.update()\n      end)\n  end\nend\n```\n\n```elixir\nChangeFavoriteQuote.to(\n  tenant,\n  \"Duke Leto Atreides\",\n  \"\"\"\n  Once you have explored a fear, it becomes less terrifying. \\\n  Part of courage comes from extending our knowledge.\n  \"\"\")\n```\n\n```elixir\nChangeFavoriteQuote.to(tenant, \"Captain America\", \"I can do this all day.\")\n```","ref":"watches.html#updating-my-favorite-quote"},{"type":"extras","title":"Takeaways - Sync Engine Part I - Single Object","doc":"The user of our web app is always kept up-to-date with the most recent favorite quote via the combined power of LiveView and our database. In some Phoenix applications, this is done via `Phoenix.PubSub`, but to do so requires setting up a separate PubSub resource and handling the messages for each kind. In this demo we instead used a feature from FoundationDB, and in doing so, we created a mini Sync Engine. We used a self-contained and generalizable approach composed of the EctoFoundationDB transaction and the LiveView server hook.\n\nThe main takeaway from this Livebook is that there is no visible dispatcher. Each process of our application can choose to register a watch, and it will be notified directly as needed from the database itself.","ref":"watches.html#takeaways"},{"type":"extras","title":"Sync Engine Part II - Collections","doc":"# Sync Engine Part II - Collections\n\n```elixir\nMix.install([\n  {:phoenix_playground, \"~> 0.1.0\"},\n  {:ecto_foundationdb, github: \"foundationdb-beam/ecto_foundationdb\"}\n])\n```","ref":"collection_syncing.html"},{"type":"extras","title":"Intro - Sync Engine Part II - Collections","doc":"Using [FoundationDB Watches](https://hexdocs.pm/ecto_foundationdb/Ecto.Adapters.FoundationDB.html#module-watches), we can set up a mini read-path Sync Engine. It will automatically propagate new data to all mounted LiveViews with push-messaging delivered from the database directly to the LiveView process.\n\nThis is Part II, where we explore syncing a collection of objects to the LiveView page. See [Part I](docs/getting_started/watches.livemd) for syncing a single object.","ref":"collection_syncing.html#intro"},{"type":"extras","title":"Setup Ecto - Sync Engine Part II - Collections","doc":"First, we set up Ecto, defining a Quote schema, and starting the Repo.\n\nThis is similar to Part I, but there's a key difference: we create a `SchemaMetadata` index on `Quote`. This metadata is required to drive watches on the collection of quotes.\n\n```elixir\ndefmodule Quote do\n  use Ecto.Schema\n\n  alias Ecto.Changeset\n\n  @primary_key {:id, :binary_id, autogenerate: true}\n\n  schema \"quotes\" do\n    field(:author, :string)\n    field(:content, :string)\n    field(:likes, :integer, default: 0)\n    field(:as_of, :naive_datetime)\n    timestamps()\n  end\n\n  def changeset(quote, params \\\\ %{}) do\n    quote\n    |> Changeset.cast(params, [:author, :content, :likes, :as_of])\n  end\n\n  def like!(tenant, id) do\n      Repo.transactional(tenant, fn ->\n        quote = Repo.get!(Quote, id)\n\n        quote\n        |> changeset(%{likes: quote.likes+1})\n        |> Repo.update!()\n      end)\n  end\nend\n\ndefmodule QuoteMetadata do\n  use EctoFoundationDB.Migration\n\n  @impl true\n  def change() do\n    [create(metadata(Quote))]\n  end\nend\n\ndefmodule Repo do\n  use Ecto.Repo, otp_app: :my_app, adapter: Ecto.Adapters.FoundationDB\n\n  use EctoFoundationDB.Migrator\n\n  @impl true\n  def migrations(), do: [\n    {1, QuoteMetadata}\n  ]\nend\n\n# ------------------------------------------------------------------\n# This section contains some set-up within Livebook that's typically\n# handled instead in your application startup and config.\nApplication.put_env(:my_app, Repo,\n  open_db: &EctoFoundationDB.Sandbox.open_db/1,\n  storage_id: EctoFoundationDB.Sandbox\n)\n\n{:ok, _} = Ecto.Adapters.FoundationDB.ensure_all_started(Repo.config(), :temporary)\nRepo.start_link(log: false)\n# ------------------------------------------------------------------\n\nalias EctoFoundationDB.Tenant\n\ntenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview-ii\")\n\n```\n\nNext, we create two `Quote`s in the DB. Our LiveView page will render the full list.\n\n```elixir\nquotes = [\n  {\"René Descartes\", \"Cogito, ergo sum.\"},\n  {\"Heraclitus\", \"One cannot step twice in the same river.\"}\n]\n\nfor {author, content} <- quotes, do:\n  Repo.insert!(%Quote{author: author, content: content}, prefix: tenant)\n```","ref":"collection_syncing.html#setup-ecto"},{"type":"extras","title":"Setup LiveView - Sync Engine Part II - Collections","doc":"Just like in Part I, we're using Phoenix Playground to create a sample LiveView. Here are the key takeways:\n\n1. In `sync_all!/4`, we're initializing our mini sync engine. Notice that we use `Repo.all` and `SchemaMetadata.watch_changes`.\n\n2. `handle_ready/2` is the same LiveView hook as Part I.\n\n3. `sync_all!/4` and `handle_ready/2` both contain nothing specific to the Quote schema. This is a general purpose approach that can be re-used for any other schema types.\n\n```elixir\ndefmodule DemoLive do\n  use Phoenix.LiveView\n\n  alias EctoFoundationDB.Indexer.SchemaMetadata\n\n  def mount(_params, _session, socket) do\n\n    tenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview-ii\")\n\n    {:ok, socket\n      |> assign(tenant: tenant)\n      |> sync_all!(Quote, :quotes)\n    }\n  end\n\n  # 1. Initialize our mini Sync Engine. Similar to Part I, except this time we\n  #    use Repo.all and SchemaMetadata.watch_changes\n  defp sync_all!(socket, schema, label) do\n    {list, futures} = Repo.transactional(socket.assigns.tenant,\n      fn ->\n        list = Repo.all(schema)\n        future = SchemaMetadata.watch_changes(schema, label: label)\n        {list, [future]}\n      end)\n\n    socket\n    |> assign(label, list)\n    |> assign(futures: futures)\n    |> attach_hook({:assign_ready, schema, label}, :handle_info, &handle_ready/2)\n  end\n\n  # 2. LiveView server hook that receives a message from the database layer\n  #    and updates the assigns (same as Part I)\n  defp handle_ready({ref, :ready}, socket) do\n\n    %{assigns: assigns} = socket\n\n    {new_assigns, futures} = Repo.assign_ready(\n        assigns.futures,\n        [ref],\n        watch?: true,\n        prefix: assigns.tenant\n      )\n    {:halt, socket\n            |> assign(new_assigns)\n            |> assign(futures: futures)}\n  end\n\n  def render(assigns) do\n    ~H\"\"\"\n     \n       A list of Quotes \n       \n         \n           \n             <%= quote.content %> \n              - <%= quote.author %>  \n\n             \n               Likes:  \n               <%= quote.likes %> \n               👍 \n             \n           \n         \n       \n     \n\n     \n      h1 { text-align: center; }\n      h3 { text-align: right; }\n      h6 { text-align: right; }\n      p { text-align: center; }\n      ul li { list-style-type: none; }\n\n      .content {\n        max-width: 500px;\n        margin: auto;\n      }\n     \n    \"\"\"\n  end\n\n  # 3. Receive the \"like\" event and update the DB\n  def handle_event(\"like\", params, socket) do\n\n    Quote.like!(socket.assigns.tenant, params[\"id\"])\n\n    {:noreply, socket}\n  end\n\nend\n\nPhoenixPlayground.start(live: DemoLive)\n```\n\nOpen a web browser to http://localhost:4000 side-by-side with this Livebook, so that you can watch the updates in real-time.\n\nYou can click the Like button and observe that the value changes. Notice that we are not pushing this value to the assigns explicitly. It's being handled automatically in our mini sync engine in `handle_info`. This is all wired together because of our use of `watch_changes` on the `Quote` schema. There are various functions that notify your app in different circumstances:\n\n* `watch_inserts`: Notifies when an insert or upsert happens\n* `watch_deletes`: Notifies when a delete happens\n* `watch_collection`: Notifies when either an insert or a delete happens\n* `watch_updates`: Notifies when an update (`Repo.update`) happens\n* `watch_changes`: Notifies when an insert, delete, or update happens\n\nAll of these notifications are scoped to the tenant.","ref":"collection_syncing.html#setup-liveview"},{"type":"extras","title":"Inserting and deleting some Quotes - Sync Engine Part II - Collections","doc":"To demonstrate that the page will update as expected, we can insert and delete some Quote objects while we have the page open.\n\n```elixir\ntenant = Tenant.open!(Repo, \"experiment-with-watches-in-liveview-ii\")\n\ndefmodule CollectQuotes do\n  def insert(tenant, author, content) do\n    Repo.insert!(%Quote{author: author, content: content}, prefix: tenant)\n  end\n\n  def delete_random(tenant) do\n    Repo.transactional(tenant, fn ->\n      Repo.all(Quote)\n      |> Enum.random()\n      |> Repo.delete!()\n    end)\n  end\nend\n```\n\n```elixir\nCollectQuotes.insert(tenant, \"Elixir\", \"Hello World\")\n```\n\n```elixir\nCollectQuotes.delete_random(tenant)\n```","ref":"collection_syncing.html#inserting-and-deleting-some-quotes"},{"type":"extras","title":"Takeaways - Sync Engine Part II - Collections","doc":"In Part I, we showed how EctoFDB can implement a mini Sync Engine for the read-path of a single object in the database. In Part II, we upgraded it to a collection of objects on the tenant.\n\nThese two approaches can be combined easily - the LiveView hook we defined works for both types at the same time. The only difference is how you initialize on mount.","ref":"collection_syncing.html#takeaways"},{"type":"extras","title":"Testing with ExUnit","doc":"# Testing with ExUnit\n\nSetting up your app to use a FoundationDB Sandbox is very easy!\n\nFirst, in `config/test.exs`:\n\n```elixir\nconfig :my_app, MyApp.Repo,\n  open_db: &EctoFoundationDB.Sandbox.open_db/1\n```\n\nNext, set up an ExUnit Case that will provide you with a new Tenant to use\nin your tests. This step is optional, but recommended: it will make writing\neach test frictionless.\n\n```elixir\ndefmodule MyApp.TenantCase do\n  @moduledoc false\n  use ExUnit.CaseTemplate\n\n  alias EctoFoundationDB.Sandbox\n\n  setup do\n    tenant_id = Ecto.UUID.autogenerate()\n\n    tenant = Sandbox.checkout(MyApp.Repo, tenant_id, [])\n\n    on_exit(fn ->\n      Sandbox.checkin(MyApp.Repo, tenant_id)\n    end)\n\n    {:ok, [tenant_id: tenant_id, tenant: tenant]}\n  end\nend\n```\n\nNow, you can use your TenantCase to do any FDB operation in a test. Because\nwe're using tenants with randomized names, you don't have to worry about key conflicts, so you can include the `async: true` option.\n\n```elixir\ndefmodule MyAppHelloFDBTest do\n  use MyApp.TenantCase, async: true\n\n  alias Ecto.Adapters.FoundationDB\n\n  test \"hello\", context do\n    tenant = context[:tenant]\n\n    # An :erlfdb hello world, delete this for your tests\n    assert :not_found ==\n             FoundationDB.transactional(\n               tenant,\n               fn tx ->\n                 tx\n                 |> :erlfdb.get(\"hello world\")\n                 |> :erlfdb.wait()\n               end\n             )\n\n    # Example:\n    #   MyApp.Repo.insert!(%MyApp.Hello{message: \"world\"}, prefix: tenant)\n  end\nend\n```","ref":"testing.html"},{"type":"extras","title":"Operator's Manual","doc":"# Operator's Manual\n\nIn order to maintain transactional guarantees as your application changes and\ngrows, there is a burden on the database operator for some data manipulations.\nEctoFoundationDB provides building blocks that\ncan be executed at strategic times to maintain data integrity.\n\nThere is a need for a human operator to execute these commands for one\nimportant reason: EctoFDB does not know details about the topology of your\ndistributed application, or the migration versions that are active on clients\nat a given time.\n\nIn writing this guide, we've taken heavy inspiration from\n[GitHub | Safe Ecto Migrations](https://github.com/fly-apps/safe-ecto-migrations)\nIt's written assuming a SQL database like Postgres, but many of the concepts\nare applicable to EctoFoundationDB, and it's worth a read.\n\n> #### Safe Index Migrations {: .note}\n> Adding and dropping indexes is done completely on-line without risk of data consistency problems.\n> The remainder of this document discusses other data operations. Please\n> refer to the documentation on\n> [Migrations](Ecto.Adapters.FoundationDB.html#module-migrations)\n> for more info on index migrations.","ref":"operators_manual.html"},{"type":"extras","title":"Safe Data Migrations - Operator's Manual","doc":"- [Adding a field](#adding-a-field)\n- [Removing a field](#removing-a-field)\n- [Renaming a field](#renaming-a-field)","ref":"operators_manual.html#safe-data-migrations"},{"type":"extras","title":"Adding a field - Operator's Manual","doc":"Adding a field is trivial. Simply modify your schema to include the new field.\nThere is no `ALTER TABLE` DDL or equivalent. Any existing objects will\nautomatically interpret values for the new field as `nil`. If you require\nsome other value, use the `Repo.update!/2` function like normal.\n\n```diff\n# App deploy, in the Ecto schema\n\ndefmodule MyApp.Post do\n  schema \"posts\" do\n+   field :a_new_field, :string\n  end\nend\n```\n\n---","ref":"operators_manual.html#adding-a-field"},{"type":"extras","title":"Removing a field - Operator's Manual","doc":"If all you care about is not having the field show up in your structs, you may\nchoose to simply remove it from your schema. The underlying data will remain\nintact, however, until each item is updated at some point in the future. If you\nneed to ensure that the data is truly removed, continue reading.\n\nSafety can be assured if the application code is first updated to remove references to the field so it's no longer loaded or queried. Then, the field can safely be removed from the database.\n\n1. Deploy code change to remove references to the field.\n2. Execute `EctoFoundationDB.CLI.delete_field!/3` to remove the data.\n\n> #### Warning: data rewrite {: .warning}\n> Removing a field requires writing all keys that have that field in the\n> value.\n\nApplication deployment:\n\n```diff\n# App deploy, in the Ecto schema\n\ndefmodule MyApp.Post do\n  schema \"posts\" do\n-   field :no_longer_needed, :string\n  end\nend\n```\n\nOnce your application is fully deployed to all nodes, execute the following\ncommand:\n\n```elixir\niex> EctoFoundationDB.CLI.delete_field!(MyApp.Repo, MyApp.Post, :no_longer_needed)\n```\n\n---","ref":"operators_manual.html#removing-a-field"},{"type":"extras","title":"Renaming a field - Operator's Manual","doc":"Take a phased approach:\n\n1. Create a new field with any relevant indexes\n2. In application code, write to both fields\n3. Backfill data from old field to new field (see below)\n4. In application code, move reads from old field to the new field\n5. In application code, remove old field from Ecto schemas, and drop any old index(es)\n6. Delete the old field (see below)\n\n> #### Warning: data rewrite {: .warning}\n> Renaming a field requires writing all related keys twice.\n\n**Backfill data from old field to new field**\n\n```elixir\niex> EctoFoundationDB.CLI.copy_field!(MyApp.Repo, MyApp.Post, :old_field, :new_field)\n```\n\n**Delete the old field**\n\n```elixir\niex> EctoFoundationDB.CLI.delete_field!(MyApp.Repo, MyApp.Post, :old_field)\n```\n\n> #### Sample Available {: .note}\n> Please refer to [GitHub | cli_test.exs](https://github.com/foundationdb-beam/ecto_foundationdb/blob/main/test/ecto/integration/cli_test.exs)\n> to review the test that executes this operation.","ref":"operators_manual.html#renaming-a-field"},{"type":"extras","title":"Metadata","doc":"# Metadata\n\nEctoFDB stores metadata alongside your Ecto schema data in order to maintain\ncorrect data integrity on the indexes. In doing so, each EctoFDB client\nknows which indexes are needed to be written to, even if the application code\nis on an older version, which can be the case when deploying updates to a\ndistributed system.\n\nSince this metadata is required to be available in every transaction, EctoFDB\nmust *allow for* the retrieval of the metadata in every transaction. However,\nwe wish to avoid that retrieval from actually happening as much as possible.\nOtherwise, those particular keys become very \"hot\" and limit the scalability\nof the system. Therefore, we maintain a local cache of the metadata for each\n(tenant, schema) tuple.\n\nThis document describes the design considerations for the metadata and the\napproaches we use to ensure the cache is guaranteed to be valid.","ref":"metadata.html"},{"type":"extras","title":"Metadata Content - Metadata","doc":"We currently keep track of 2 types of metadata: indexes and partial indexes.","ref":"metadata.html#metadata-content"},{"type":"extras","title":"Indexes - Metadata","doc":"This is a list of all 'ready' indexes for a particular Schema. An index is\nconsidered 'ready' when it has been created and is ready to be used.\n\nAt query time, this list of indexes is inspected to determine the index to use\nto provide the best possible query performance.\n\nAt insert and update time, the list is used to set or clear the index keys\naccording to the specified `EctoFoundationDB.Indexer`.","ref":"metadata.html#indexes"},{"type":"extras","title":"Partial Indexes - Metadata","doc":"This is a list of all indexes that are currently undergoing a migration,\nusually for index creation. The creation of an index for a particular\n(tenant, schema) tuple can take arbitrarily long, and so any queries that\narrive in the meantime must be handled accordingly. The list of partial\nindexes informs EctoFDB of how to keep data integrity for any concurrent\ncreation of `Default` indexes.","ref":"metadata.html#partial-indexes"},{"type":"extras","title":"Metadata Cache - Metadata","doc":"The cache uses 2 stages for invalidation.\n\n* **Stage 1:** The global FoundationDB metadata version key\n* **Stage 2:** A version key for each (tenant, schema) tuple, and a special\n\"claim key\" for any partial indexes\n\n#### Metadata Version Key\n\nThis is a key that has special treatment in the implementation of\nFoundationDB transactions. The\n[FDB Design Doc | Metadata Version](https://github.com/apple/foundationdb/blob/deda04b8453ecbc6411cc7ac41efb3213e18343f/design/metadata-version.md)\nprovides a detailed explanation of how this key is implemented.\n\nFor the purposes of this document, it's important to understand that the metadata version key is truly global to the entire keyspace. Since we allow\ntenants to migrate independently, this necessarily means that the migration for\na single tenant will invalidate Stage 1 of the cache for **all tenants** in\nthe  database.\n\nThere is no cost to reading the global metadata version key, since it's\nalways sent along with other necessary transactional data. This means that\nat steady state, your transactions only need to do the base minimum FDB\noperations, and are able to avoid any hot keys.\n\nIf the global version is found to have changed, we do not yet invalidate the\ncache. Instead, we move onto Stage 2 of the cache invalidation process. In\ndoing so, we limit the impact of another tenant's migration.\n\n#### Schema Migration Version Key\n\nThis is a key that is managed by EctoFDB itself, and does not have special\ntreatment by FDB. It always contains an integer value representing the largest\nversion number for complete migrations that are specified in your\n`EctoFoundationDB.Migrator`.\n\nWhen the global key is found to have changed in Stage 1 of the cache\ninvalidation, we perform a 'get' on this key. Then we delay the wait until\nafter some of the transaction work has been completed. Thus, the cache\nis *optimistic* that the cached value is still valid. Only at the end of the\ntransaction do we wait for and compare the version value. If it's not equal,\nthen the cache is invalidated and the transaction is retried.\n\n#### Claim Key\n\nThis key is used for tracking metadata that is currently undergoing a\nmigration. It's managed in the same way as the schema migration version key.\nIts content includes a cursor that defines the progress of the migration.","ref":"metadata.html#metadata-cache"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}